/*
  G. Wiesenekker. GWD a draughts program.
  Copyright (C) 1993-1994  G. Wiesenekker
  E-mail: wiesenekker@sara.nl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _DRAUGHTS_H
#define _DRAUGHTS_H

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <signal.h>

#define MAC    0
#define DOS    1
#define RS6000 2
#define CRAY   3
#define SGI    4
#define SUN    5

#ifndef MACHINE
#ifdef __MSDOS__
#define MACHINE DOS
#include <alloc.h>
#else
#define MACHINE MAC
#endif
#endif

#if MACHINE == SUN
#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1
#define CLOCKS_PER_SEC 100000L
#define difftime(X, Y) ((X) - (Y))
#endif

#undef SMALL
#undef GUI

#define DEBUG_LEVEL 0
#ifdef GUI
#define OUTPUT 0
#else
#define OUTPUT 1
#endif
#undef NULL_MOVE
#undef THREAT

#define NULL_MOVE_DEPTH 3
#define NULL_MOVE_REDUC 1

#define THREAT_DEPTH 3
#define THREAT_REDUC 1
#define THREAT_MARGIN (2 * MAN_SCORE)
#define NTHREAT_MAX 2

extern int DepthMax;
extern int TimeLimit;
extern int TimeExtra;
extern int Append;
extern int GameTimeUsed;
extern int UsePos;
extern int UseEndgame;

#define FALSE 0
#define TRUE 1
#define INVALID (-1)
#define LINE_MAX 256

#define and &&
#define or  ||

#define ABS(A) ((A) > 0 ? (A) : -(A))
#define MAX(A, B) ((A) > (B) ? (A) : (B))
#define BIT(X)  (1 << (X))
#define BITL(X) (1L << (X))

#ifndef FATAL_ERROR
#ifdef SMALL
#define FATAL_ERROR(X)\
	{my_printf("??%s::%lu\n", FILE_NAME, (ulong_t) __LINE__); exit(EXIT_FAILURE);}
#else
#define FATAL_ERROR(X)\
	{my_printf("%s::%lu:%s\n", FILE_NAME, (ulong_t) __LINE__, X); exit(EXIT_FAILURE);}
#endif
#endif

#ifndef BUG
#define BUG(X) {if (X) FATAL_ERROR(#X)}
#endif

#if DEBUG_LEVEL == 0
#define RANGE(I, N)
#define PRANGE(P, B, N)
#else
#define RANGE(I, N) {if (((I) < 0) or ((I) >= (N))) FATAL_ERROR("range!");}
#define PRANGE(P, B, N) {if (((P) == NULL) or (((P) - (B)) >= (N))) FATAL_ERROR("range!");}
#endif

#define FCLOSE(F) BUG(fclose(F) == EOF)
#if MACHINE == DOS
#define MALLOC(P, T, N)\
{\
	my_printf("malloc %s size=%lu number=%lu total=%lu\n", #P, (ulong_t) sizeof(T), (ulong_t) (N), (ulong_t) sizeof(T) * (N));\
	BUG(((P) = (T *) malloc(sizeof(T) * (N))) == NULL)\
}
#define FREE(P) {free(P);}
#else
#define MALLOC(P, T, N) BUG(((P) = (T *) malloc(sizeof(T) * (N))) == NULL)
#define FREE(P) {free(P);}
#endif
#if MACHINE == DOS
#define HUGE_MALLOC(P, T, N)\
{\
	my_printf("huge malloc %s size=%lu number=%lu total=%lu\n", #P, (ulong_t) sizeof(T), (ulong_t) (N), (ulong_t) sizeof(T) * (N));\
	BUG(((P) = (T huge *) farmalloc(sizeof(T) * (N))) == NULL)\
}
#define HUGE_FREE(P) {farfree((void far *) P);}
#define HUGE_FREAD(P, T, N, F)\
{\
	long i;\
	T t;\
	for (i = 0; i < (N); i++)\
	{\
		BUG(fread(&t, sizeof(T), 1, F) != 1)\
		(P)[i] = t;\
	}\
}
#define HUGE_FWRITE(P, T, N, F)\
{\
	long i;\
	T t;\
	for (i = 0; i < (N); i++)\
	{\
		t = (P)[i];\
		BUG(fwrite(&t, sizeof(T), 1, F) != 1)\
	}\
}
#define XMS_FREAD(P, T, N, F)\
{\
	long i;\
	T t;\
	for (i = 0; i < (N); i++)\
	{\
		BUG(fread(&t, sizeof(T), 1, F) != 1)\
		XMShandle = P;\
		XMSoffset = i;\
		put_endgame(t);\
	}\
}
#define XMS_FWRITE(P, T, N, F)\
{\
	long i;\
	T t;\
	for (i = 0; i < (N); i++)\
	{\
		XMShandle = P;
		XMSoffset = i;
		t = get_endgame();\
		BUG(fwrite(&t, sizeof(T), 1, F) != 1)\
	}\
}
#endif

typedef signed char byte_t;
typedef unsigned int uint_t;
#if MACHINE != SGI
typedef unsigned long ulong_t;
#endif

#define DUMMY(F, V) if (dummy) my_printf(F, V);
extern int dummy;

enum {WHITE_COLOUR = 0, BLACK_COLOUR, NONE, SIDE};

extern int my_colour, your_colour;
extern char *colour_as_string[2];

enum {MAN = 0, CROWN};
#define MAN_SCORE   (300)
#define CROWN_SCORE (1000)

typedef int score_t;

#if MACHINE == DOS
typedef byte_t move_t;
#else
typedef int move_t;
#endif

typedef struct
{
	ulong_t lo, hi;
} hash_key_t;

#define WBIT  0
#define BBIT  1
#define MBIT  2
#define CBIT  3
#define WMBIT 4
#define BMBIT 5
#define WCBIT 6
#define BCBIT 7
#define W(X)     ((b[X] & BIT(WBIT)) != 0)
#define NOTW(X)  !(W(X))
#define B(X)     ((b[X] & BIT(BBIT)) != 0)
#define NOTB(X)  !(B(X))
#define M(X)     ((b[X] & BIT(MBIT)) != 0)
#define NOTM(X)  !M(X)
#define C(X)     ((b[X] & BIT(CBIT)) != 0)
#define NOTC(X)  !C(X)
#define WM(X)    ((b[X] & BIT(WMBIT)) != 0)
#define NOTWM(X) !WM(X)
#define BM(X)    ((b[X] & BIT(BMBIT)) != 0)
#define NOTBM(X) !BM(X)
#define WC(X)    ((b[X] & BIT(WCBIT)) != 0)
#define NOTWC(X) !WC(X)
#define BC(X)    ((b[X] & BIT(BCBIT)) != 0)
#define NOTBC(X) !BC(X)
#define EMP(X)   (b[X] == 0)
#define ANY(X)   (b[X] != 0)
#define WTOT  (tot_count[WHITE_COLOUR])
#define BTOT  (tot_count[BLACK_COLOUR])
#define WMTOT (piece_count[WHITE_COLOUR][MAN])
#define BMTOT (piece_count[BLACK_COLOUR][MAN])
#define WCTOT (piece_count[WHITE_COLOUR][CROWN])
#define BCTOT (piece_count[BLACK_COLOUR][CROWN])

#define BOARD_MAX 66

extern int pat_map[BOARD_MAX];
typedef byte_t pat_t;
extern pat_t b[51];

typedef struct
{
	int board_colour, board_kind, board_value, board_where;
	hash_key_t *board_key;
	int *board_dir;
	int *board_count;
	pat_t board_pat;
} board_t;

#define LIST_MAX 21

typedef struct
{
	move_t **move_base, *stack_base;
	board_t **capt_base, *prom;
	hash_key_t node_key, node_total_key;
	int node_mat[2];
	int node_reverse;
	move_t *node_move;
} node_t;

extern int white_piece_count[2];
extern int black_piece_count[2];
extern int *piece_count[2];
extern int tot_count[2];
extern node_t *node;

#define STRING_MAX 64

typedef struct
{
	int root_move;
	score_t root_score;
	int root_hit;
} root_t;

typedef int var_t;

typedef byte_t endgame_t;

#define NROOT_MAX 128

#define NGAME_MAX 1024
extern char *game_moves;

extern int nroot;
extern root_t *root;
extern int jroot;
extern int GameTimeUsed;

#define ENDGAME_INVAL 101
#define ENDGAME_DRAW  100

#ifdef __cplusplus
extern "C" {
#endif

void my_printf(char *, ...);
ulong_t ulrand(void);
void handle_signal(int);
extern char move_string[STRING_MAX];
void move_to_string(move_t *, int);
int read_board(FILE *);
void display_board(void);
void write_board(FILE *f);
void display_score(void);
void display_list(void);
void compact_lists(void);
void update_key(hash_key_t *);
void do_move(move_t *);
void undo_move(void);
void set_score(double *);
int draw_by_repetition(int);
void gen_root(void);
int search_root(char *, int);
void do_root(void);
void sort_root(void);
int return_pat_score(void);

void analyze_pos(char *);
void analyze_game(char *);
double solve_problems(char *);
void play_batch(char *);
void play_game(int, int, char *);
void optimise(char *);
void init(void);
void quit(void);

int return_endgame_score(int, int);
void do_endgame(void);
void minimax(void);
#ifdef GUI
move_t *get_move(int);
int check_time(void);
void print_cur_move(int, int);
void print_best_move(int, int);
void print_cache(char, var_t *);
void print_cache_move(char *);
void gwdupdatetime(void);
void gwdclearpv(void);
void gwdaddpv(char *);
void gwddrawpv(void);
void gwddrawcurrentmove(char *);
void gwddrawbestmove(char *);
int gwdabortthink(void);
#endif

#ifdef __cplusplus
}
#endif

#endif
