/*
  G. Wiesenekker. GWD a draughts program.
  Copyright (C) 1993-1994 G. Wiesenekker
  E-mail: wiesenekker@sara.nl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "draughts.h"
static char *FILE_NAME = __FILE__;

void my_printf(char *format, ...)
{
	va_list arg;
	static FILE *stdecho = NULL;
	
	if (strcmp(format, "%z") == 0)
	{
		if (stdecho != NULL) fclose(stdecho);
		return;
	}
	va_start(arg, format);
#if OUTPUT > 0
	vfprintf(stdout, format, arg);
	fflush(stdout);
#endif
	if (stdecho == NULL)
	{
		BUG((stdecho = fopen("echo", "w")) == NULL)
		fprintf(stdecho, "run at ??\n");
		fflush(stdecho);
	}
	vfprintf(stdecho, format, arg);
	fflush(stdecho);
	va_end(arg);
}

ulong_t ulrand(void)
{
	/*
	random numbers from Mathematica 2.0.
	SeedRandom = 1;
	Table[Random[Integer, {0, 2^32 - 1}]
	*/
	static ulong_t x[55] =
	{
		1410651636UL,
		3012776752UL,
		3497475623UL,
		2892145026UL,
		1571949714UL,
		3253082284UL,
		3489895018UL,
		387949491UL, 
		2597396737UL,
		1981903553UL,
		3160251843UL,
		129444464UL, 
		1851443344UL,
		4156445905UL,
		224604922UL,
		1455067070UL, 
		3953493484UL,
		1460937157UL,
		2528362617UL,
		317430674UL, 
		3229354360UL,
		117491133UL,
		832845075UL,
		1961600170UL, 
		1321557429UL,
		747750121UL,
		545747446UL,
		810476036UL,
		503334515UL, 
		4088144633UL,
		2824216555UL,
		3738252341UL,
		3493754131UL, 
		3672533954UL,
		29494241UL,
		1180928407UL,
		4213624418UL,
		33062851UL, 
		3221315737UL,
		1145213552UL,
		2957984897UL,
		4078668503UL, 
		2262661702UL,
		65478801UL,
		2527208841UL,
		1960622036UL,
		315685891UL, 
		1196037864UL,
		804614524UL,
		1421733266UL,
		2017105031UL, 
		3882325900UL,
		810735053UL,
		384606609UL,
		2393861397UL
	};
	static int init = TRUE;
	static ulong_t y[55];
	static int j, k;
	ulong_t ul;
	
	if (init)
	{
		int i;
		
		init = FALSE;
		for (i = 0; i < 55; i++)
			y[i] = x[i];
		j = 24 - 1;
		k = 55 - 1;
	}
	
	ul = (y[k] += y[j]);
	if (--j < 0) j = 55 - 1;
	if (--k < 0) k = 55 - 1;
	return(ul);
}

void handle_signal(int sig)
{
	static int nint = 0;
	
	if (sig == SIGINT)
	{
		nint++;
		if (nint < 2)
		{
			my_printf("caught SIGINT %d!\n", nint);
			BUG(signal(SIGINT, handle_signal) == SIG_ERR);
		}
		else
		{
			my_printf("SIGINT!\n");
			exit(EXIT_FAILURE);
		}
	}
}

#define START_BOARD "brd.txt"

#ifndef GUI
void store_game(FILE *);
void store_game(FILE *f)
{
	char line[LINE_MAX];
	char move[LINE_MAX];
	
	strcpy(game_moves, "");
	while(fgets(line, LINE_MAX, f) != NULL)
	{
		if (sscanf(line, "%s", move) != 1) continue;
		strcat(game_moves, move);
		strcat(game_moves, "\n");
	}
	my_printf("%s", game_moves);
}

int read_game(void);
int read_game(void)
{
	char move[LINE_MAX];
	int iroot;
	char *p;
	int nwhite, nblack;
	
	p = game_moves;
	my_colour = WHITE_COLOUR;
	your_colour = BLACK_COLOUR;
	nwhite = nblack = 0;
	while(sscanf(p, "%s", move) == 1)
	{
		gen_root();
		BUG((iroot = search_root(move, TRUE)) == INVALID)
		if (my_colour == WHITE_COLOUR)
			nwhite++;
		else
			nblack++;
		do_move(node->move_base[root[iroot].root_move]);
		p += strlen(move) + 1;
	}
	compact_lists();
	if (my_colour == WHITE_COLOUR)
		return(nwhite);
	else
		return(nblack);
}

void analyze_pos(char *name)
{
	FILE *f;
	int iroot;
	
	BUG((f = fopen(name, "r")) == NULL)
	while(read_board(f))
	{
		my_printf("pos board\n");
		display_board();
		display_list();
		display_score();
		my_colour = WHITE_COLOUR;
		your_colour = BLACK_COLOUR;
		gen_root();
		my_printf("white moves:\n");
		for (iroot = 0; iroot < nroot; iroot++)
		{
			do_move(node->move_base[root[iroot].root_move]);
			undo_move();
			move_to_string(node->move_base[root[iroot].root_move], TRUE);
			my_printf(" %s", move_string);
			if (((iroot % 10) == 9) or (iroot == (nroot - 1))) my_printf("\n");
		}
		my_printf("black moves:\n");
		my_colour = BLACK_COLOUR;
		your_colour = WHITE_COLOUR;
		gen_root();
		for (iroot = 0; iroot < nroot; iroot++)
		{
			do_move(node->move_base[root[iroot].root_move]);
			undo_move();
			move_to_string(node->move_base[root[iroot].root_move], TRUE);
			my_printf(" %s", move_string);
			if (((iroot % 10) == 9) or (iroot == (nroot - 1))) my_printf("\n");
		}
	}
	BUG(fclose(f) == EOF)
}

void analyze_game(char *name)
{
	FILE *f;
	char line[LINE_MAX];
	char move[LINE_MAX];
	int iroot;
	
	BUG((f = fopen(START_BOARD, "r")) == NULL)
	BUG(!read_board(f))
	BUG(fclose(f) == EOF)
	display_board();
	display_score();
	
	BUG((f = fopen(name, "r")) == NULL)
	while(fgets(line, LINE_MAX, f) != NULL)
	{
		if (sscanf(line, "%s", move) != 1) continue;
		gen_root();
		BUG((iroot = search_root(move, TRUE)) == INVALID)
		do_move(node->move_base[root[iroot].root_move]);
		display_board();
		display_score();
	}
	BUG(fclose(f) == EOF)
}

double solve_problems(char *name)
{
	FILE *f;
	int nproblems, nsolved, nsimple, ninvalid;
	char col_to_move[LINE_MAX], move[LINE_MAX];
	int iroot;
	
	nproblems = nsolved = nsimple = ninvalid = 0;
	BUG((f = fopen(name, "r")) == NULL)
	while(read_board(f))
	{
		BUG(fscanf(f, "%s%s", col_to_move, move) != 2)
		nproblems++;
		display_board();
		display_list();
		
		node->node_total_key = node->node_key;
		if (*col_to_move == 'W')
		{
			my_colour = WHITE_COLOUR;
			your_colour = BLACK_COLOUR;
		}
		else
		{
			my_colour = BLACK_COLOUR;
			your_colour = WHITE_COLOUR;
		}
		update_key(&(node->node_total_key));
		gen_root();
		if (nroot == 0)
		{
			my_printf("problem %d is draw or mate!\n", nproblems);
			nsimple++;
			continue;
		}
		iroot = search_root(move, TRUE);
		if (iroot == INVALID)
		{
			my_printf("problem %d has an invalid move!\n", nproblems);
			ninvalid++;
			continue;
		}
		if (nroot == 1)
		{
			my_printf("problem %d has a forced move!\n", nproblems);
			nsimple++;
		}
		do_root();
		if (iroot == jroot)
		{
			my_printf("solved problem %d!\n", nproblems);
			nsolved++;
		}
		else
			my_printf("did not solve problem %d!\n", nproblems);
	}
	my_printf("nproblems=%d nsolved=%d nsimple=%d ninvalid=%d\n",
		nproblems, nsolved, nsimple, ninvalid);
	BUG(fclose(f) == EOF)
	if (nproblems == 0)
		return((double) 0);
	else
		return((double) nsolved / (double) nproblems);
}

void solve_endgame(char *);
void solve_endgame(char *name)
{
	FILE *f;
	char col_to_move[LINE_MAX];
	
	BUG((f = fopen(name, "r")) == NULL)
	while(read_board(f))
	{
		BUG(fscanf(f, "%s", col_to_move) != 1)
		display_board();
		display_list();
		
		if (*col_to_move == 'W')
		{
			my_colour = WHITE_COLOUR;
			your_colour = BLACK_COLOUR;
		}
		else
		{
			my_colour = BLACK_COLOUR;
			your_colour = WHITE_COLOUR;
		}
		node->node_total_key = node->node_key;
		update_key(&(node->node_total_key));
		minimax();
	}
	BUG(fclose(f) == EOF)
}

void play_batch(char *name)
{
	FILE *f;
	
	BUG((f = fopen(START_BOARD, "r")) == NULL)
	BUG(!read_board(f))
	BUG(fclose(f) == EOF)
	
	BUG((f = fopen(name, "r")) == NULL)
	store_game(f);
	read_game();
	BUG(fclose(f) == EOF)
	my_printf("board after read game\n");
	display_board();
	display_list();
	display_score();
	
	if (draw_by_repetition(TRUE))
	{
		my_printf("draw by rep!\n");
		exit(EXIT_FAILURE);
	}
	
	gen_root();
	if (nroot == 0)
	{
		my_printf("%sis mate!\n", colour_as_string[my_colour]);
		exit(EXIT_FAILURE);
	}
	if (nroot == 1)
	{
		my_printf("forced!\n");
		jroot = 0;
	}
	else
		do_root();
	if (Append)
	{
		BUG((f = fopen(name, "a")) == NULL)
		move_to_string(node->move_base[root[jroot].root_move], TRUE);
		fprintf(f, "%s gwd %.3f\n", move_string, (double) root[jroot].root_score);
		BUG(fclose(f) == EOF)
	}
}

#if MACHINE == MAC
#define GAME_TIME1 (10 * 60)
#define GAME_TIME2 (5 * 60)
#define NGAME_MOVES1 20
#define NGAME_MOVES2 40
#define GAME_MARGIN (1 * 60)
#else
#define GAME_TIME1 (90 * 60)
#define GAME_TIME2 (60 * 60)
#define NGAME_MOVES1 60
#define NGAME_MOVES2 40
#define GAME_MARGIN (5 * 60)
#endif

void play_game(int xxxxxx, int yyyyyy, char *name)
{
	FILE *f;
	time_t wall1, wall2;
	int wall;
	int iroot;
	int ngame_moves;
	
	DUMMY("%d", yyyyyy)
	while(TRUE)
	{
		BUG((f = fopen(START_BOARD, "r")) == NULL)
		BUG(!read_board(f))
		BUG(fclose(f) == EOF)
		
		BUG((f = fopen(name, "r")) == NULL)
		store_game(f);
		ngame_moves = read_game();
		BUG(fclose(f) == EOF)
		my_printf("board after read game\n");
		display_board();
		/*
		display_list();
		display_occup();
		display_score();
		*/
	
		if (draw_by_repetition(TRUE))
		{
			my_printf("draw by rep!\n");
			exit(EXIT_FAILURE);
		}
		
		gen_root();
		if (nroot == 0)
		{
			my_printf("%sis mate!\n", colour_as_string[my_colour]);
			exit(EXIT_FAILURE);
		}
		else
		{
			if (xxxxxx == my_colour)
			{
				if (nroot == 1)
				{
					my_printf("forced!\n");
					jroot = 0;
					root[jroot].root_score = 0;
				}
				else
				{
					if (ngame_moves < NGAME_MOVES1)
					{
						TimeLimit =
						 	(GAME_TIME1 - GAME_MARGIN - GameTimeUsed) /
						 	(NGAME_MOVES1 - ngame_moves);
					}
					else if (ngame_moves < (NGAME_MOVES1 + NGAME_MOVES2))
					{
						TimeLimit =
							(GAME_TIME1 + GAME_TIME2 - GAME_MARGIN - GameTimeUsed) /
						  (NGAME_MOVES1 + NGAME_MOVES2 - ngame_moves);
					}
					else
					{
						TimeLimit = 90;
					}
					if (TimeLimit < 0)
					{
						my_printf("warning: negative time limit!\n");
						TimeLimit = 1;
					}
					my_printf("TimeLimit=%d\n", TimeLimit);
					wall1 = time(NULL);
					do_root();
					wall2 = time(NULL);
					wall = (int) difftime(wall2, wall1) + 1;
					GameTimeUsed += wall;
					my_printf("WallTime=%d GameTimeUsed=%d\n", wall, GameTimeUsed);
				}
				move_to_string(node->move_base[root[jroot].root_move], TRUE);
				my_printf("\agwd move: %s\n", move_string);
				BUG((f = fopen(name, "a")) == NULL)
				fprintf(f, "%s gwd %.3f\n", move_string, (double) root[jroot].root_score);
				BUG(fclose(f) == EOF)
			}
			else
			{
				char line[LINE_MAX], move[LINE_MAX];
				
				if ((ngame_moves > 0) and (ngame_moves % 5) == 0)
				{
					while(TRUE)
					{
						my_printf("\aGameTimeUsed? (min) ");
						if (gets(line) == NULL) exit(EXIT_FAILURE);
						if (sscanf(line, "%d", &GameTimeUsed) != 1) continue;
						GameTimeUsed *= 60;
						break;
					}
				}
				while(TRUE)
				{
					my_printf("your move? ");
					if (gets(line) == NULL) exit(EXIT_FAILURE);
					if (sscanf(line, "%s", move) != 1) continue;
					if (strcmp(move, "quit") == 0) exit(EXIT_SUCCESS);
					if (strcmp(move, "?") == 0)
					{
						for (iroot = 0; iroot < nroot; iroot++)
						{
							move_to_string(node->move_base[root[iroot].root_move], TRUE);
							my_printf(" %5s", move_string);
							if (((iroot % 10) == 9) or (iroot == (nroot - 1))) my_printf("\n");
						}
						continue;
					}
					iroot = search_root(move, FALSE);
					if (iroot == INVALID) continue;
					move_to_string(node->move_base[root[iroot].root_move], TRUE);
					BUG((f = fopen(name, "a")) == NULL)
					fprintf(f, "%s\n", move_string);
					BUG(fclose(f) == EOF)
					break;
				}
			}
		}
	}
}

#define NDIM 23
#define PROBLEMS_FILE "opt.txt"
#define INIT_FILE "init.txt"
#define RESULTS_FILE "results.txt"

void try(void);
void try(void)
{
	FILE *f, *g;
	double err;
	int i;
	double y[NDIM];
	
	BUG((f = fopen(INIT_FILE, "r")) == NULL)
	BUG((g = fopen(RESULTS_FILE, "a")) == NULL)
	while(fscanf(f, "%lf", &err) == 1)
	{
		BUG(err > 1.0)
		for (i = 0; i < NDIM; i++)
			BUG(fscanf(f, "%lf", y + i) != 1)
		if (err < 0)
		{
			set_score(y);
			err = solve_problems(PROBLEMS_FILE);
			fprintf(g, "%.2f", err);
			for (i = 0; i < NDIM; i++)
				fprintf(g, " %d", (int) y[i]);
			fprintf(g, "\n");
		}
	}
	BUG(fclose(g) == EOF)
	BUG(fclose(f) == EOF)
}

#if MACHINE == MAC
#define ARGC_MAX 256
int argc;
char *argv_table[ARGC_MAX];
char **argv = argv_table;
int main(void);
int main(void)
#else
int main(int argc, char **argv);
int main(int argc, char **argv)
#endif
{
#if MACHINE == MAC
	{
		FILE *f;
		char line[LINE_MAX];
	
		BUG((f = fopen("arg.txt", "r")) == NULL)
		argc = 0;
		while (fscanf(f, "%s", line) == 1)
		{
			BUG(argc >= ARGC_MAX)
			MALLOC(argv_table[argc], char, strlen(line) + 1)
			strcpy(argv_table[argc], line);
			argc++;
		}
		BUG(fclose(f) == EOF)
	}
#else
	argc--;
	argv++;
#endif

	signal(SIGINT, handle_signal);
	
	while((argc > 0) and (*argv[0] == '-'))
	{
		if (strcmp(argv[0], "-DepthMax") == 0)
		{
			BUG(argc < 2)
			BUG(sscanf(argv[1], "%d", &DepthMax) != 1)
			argc--; argc--;
			argv++; argv++;
		}
		else if (strcmp(argv[0], "-TimeLimit") == 0)
		{
			BUG(argc < 2)
			BUG(sscanf(argv[1], "%d", &TimeLimit) != 1)
			argc--; argc--;
			argv++; argv++;
		}
		else if (strcmp(argv[0], "-GameTimeUsed") == 0)
		{
			BUG(argc < 2)
			BUG(sscanf(argv[1], "%d", &GameTimeUsed) != 1)
			BUG(GameTimeUsed < 1)
			argc--; argc--;
			argv++; argv++;
		}
		else if (strcmp(argv[0], "-NoAppend") == 0)
		{
			Append = FALSE;
			argc--;
			argv++;
		}
		else if (strcmp(argv[0], "-NoPos") == 0)
		{
			UsePos = FALSE;
			argc--;
			argv++;
		}
		else if (strcmp(argv[0], "-NoEndgame") == 0)
		{
			UseEndgame = FALSE;
			argc--;
			argv++;
		}
		else
		{
			my_printf("unknown option %s ignored\n", argv[0]);
			argc--;
			argv++;
		}
	}
	
	my_printf("DepthMax=%d TimeLimit=%d GameTimeUsed=%d UsePos=%d UseEndgame=%d\n",
		DepthMax, TimeLimit, GameTimeUsed, UsePos, UseEndgame);
	GameTimeUsed *= 60;
	
	BUG(argc < 1)
	
	init();
	
	if (strcmp(argv[0], "pos") == 0)
	{
		BUG(argc < 2)
		analyze_pos(argv[1]);
	}
	else if (strcmp(argv[0], "anal") == 0)
	{
		BUG(argc < 2)
		analyze_game(argv[1]);
	}
	else if (strcmp(argv[0], "solve") == 0)
	{
		BUG(argc < 2)
		solve_problems(argv[1]);
	}
	else if (strcmp(argv[0], "batch") == 0)
	{
		BUG(argc < 2)
		play_batch(argv[1]);
	}
	else if (strcmp(argv[0], "white") == 0)
	{
		BUG(argc < 2)
		play_game(WHITE_COLOUR, BLACK_COLOUR, argv[1]);
	}
	else if (strcmp(argv[0], "black") == 0)
	{
		BUG(argc < 2)
		play_game(BLACK_COLOUR, WHITE_COLOUR, argv[1]);
	}
	else if (strcmp(argv[0], "try") == 0)
	{
		BUG(argc < 1)
		try();
	}
	else if (strcmp(argv[0], "gen") == 0)
	{
		do_endgame();
	}
	else if (strcmp(argv[0], "endgame") == 0)
	{
		BUG(argc < 2)
		solve_endgame(argv[1]);
	}
	else
	{
		my_printf("unknown arg %s!\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	quit();
	return(EXIT_SUCCESS);
}
#endif

