/*
	G. Wiesenekker. ZZZZZZ a chess program.
	Copyright (C) 1993-1994 G. Wiesenekker
	E-mail: wiesenekker@sara.nl

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "zzzzzz.h"
local char *FILE_NAME = __FILE__;

#ifndef DEBUG_LEVEL
	#define DEBUG_LEVEL 0
#endif

void util_version(char *version_string)
{
	sprintf(version_string, "%s %s %d", __DATE__, __TIME__, DEBUG_LEVEL);
}

char *time_stamp(void)
{	
	time_t t;
	
	t = time(NULL);
	return(ctime(&t));
}

/*
my_printf works just like printf, but optionally echoes to a file
(and flushes the buffer, so that you can still see what went wrong
just before the core dump)
*/

void my_printf(char *format, ...)
{
	va_list arg;
#if MACHINE == MAC
	static FILE *stdecho = NULL;
#endif
#ifdef PVM
	static FILE *pvmecho = NULL;
	char *name;
#endif
	
	va_start(arg, format);
	vfprintf(stdout, format, arg);
	fflush(stdout);
#if MACHINE == MAC
	if (stdecho == NULL)
	{
		BUG((stdecho = fopen("echo", "w")) == NULL)
		fprintf(stdecho, "Logfile echo opened at %s\n", time_stamp());
	}
	vfprintf(stdecho, format, arg);
	fflush(stdecho);
#endif
#ifdef PVM
	if (pvmecho == NULL)
	{
		if (strcmp(format, "%z") == 0)
		{
			name = va_arg(arg, char *);
			BUG((pvmecho = fopen(name, "w")) == NULL)
			fprintf(pvmecho, "Logfile %s opened at %s\n", name, time_stamp());
			fflush(pvmecho);
		}
	}
	else
	{
		vfprintf(pvmecho, format, arg);
		fflush(pvmecho);
	}
#endif
	va_end(arg);
}

/*
my_fgets does an fgets, but removes the newline.
*/

char *my_fgets(char *s, int n, FILE *stream)
{
	char *p;
	
	if (fgets(s, n, stream) == NULL)
		return(NULL);
	for (p = s; *p != '\0'; p++)
		if (*p == '\n') *p = '\0';
	return(s);
}

/*
A 32 bit random number generator. An implementation in C of the algorithm given by
Knuth, the art of computer programming, vol. 2, pp. 26-27. We use e=32, so 
we have to evaluate y(n) = y(n - 24) + y(n - 55) mod 2^32, which is implicitly
done by unsigned arithmetic.
*/

ulong_t ulrand(void)
{
	/*
	random numbers from Mathematica 2.0.
	SeedRandom = 1;
	Table[Random[Integer, {0, 2^32 - 1}]
	*/
	static ulong_t x[55] =
	{
		1410651636UL,
		3012776752UL,
		3497475623UL,
		2892145026UL,
		1571949714UL,
		3253082284UL,
		3489895018UL,
		387949491UL, 
		2597396737UL,
		1981903553UL,
		3160251843UL,
		129444464UL, 
		1851443344UL,
		4156445905UL,
		224604922UL,
		1455067070UL, 
		3953493484UL,
		1460937157UL,
		2528362617UL,
		317430674UL, 
		3229354360UL,
		117491133UL,
		832845075UL,
		1961600170UL, 
		1321557429UL,
		747750121UL,
		545747446UL,
		810476036UL,
		503334515UL, 
		4088144633UL,
		2824216555UL,
		3738252341UL,
		3493754131UL, 
		3672533954UL,
		29494241UL,
		1180928407UL,
		4213624418UL,
		33062851UL, 
		3221315737UL,
		1145213552UL,
		2957984897UL,
		4078668503UL, 
		2262661702UL,
		65478801UL,
		2527208841UL,
		1960622036UL,
		315685891UL, 
		1196037864UL,
		804614524UL,
		1421733266UL,
		2017105031UL, 
		3882325900UL,
		810735053UL,
		384606609UL,
		2393861397UL
	};
	static int init = TRUE;
	static ulong_t y[55];
	static int j, k;
	ulong_t ul;
	
	if (init)
	{
		int i;
		
		init = FALSE;
		for (i = 0; i < 55; i++)
			y[i] = x[i];
		j = 24 - 1;
		k = 55 - 1;
	}
	
	ul = (y[k] += y[j]);
	if (--j < 0) j = 55 - 1;
	if (--k < 0) k = 55 - 1;
	return(ul);
}

local void gen_crc(void);
local void gen_crc(void)
{
	ulong_t r[8];
	ulong_t m;
	int i, j;

  r[0] = 
		(BITUL(26)|BITUL(23)|BITUL(22)|BITUL(16)|BITUL(12)|BITUL(11)|BITUL(10)
		|BITUL(8) |BITUL(7) |BITUL(5) |BITUL(4) |BITUL(2) |BITUL(1) |BITUL(0));
	for (i = 1; i < 8; i++)
		r[i] = (r[i - 1] & BITUL(31)) ? (r[i - 1] << 1) ^ r[0] : r[i - 1] << 1;
	my_printf ("static ulong_t crctab[256] =\n{\n");
	for (j = 0; j < 256; j++)
	{
		m = 0;
		for (i = 0; i < 8; i++)
			if (BITUL(i) & j) m = m ^ r[i];
		my_printf ("\t%luUL,\n", m);
	}
  my_printf ("};\n");
}

/*
The GNU cksum routine from the textutils distribution, but
without the length and the one's complement.
*/

ulong_t crc(uchar_t *u)
{
	ulong_t crc;
	static ulong_t crctab[256] =
	{
		0UL,
		79764919UL,
		159529838UL,
		222504665UL,
		319059676UL,
		398814059UL,
		445009330UL,
		507990021UL,
		638119352UL,
		583659535UL,
		797628118UL,
		726387553UL,
		890018660UL,
		835552979UL,
		1015980042UL,
		944750013UL,
		1276238704UL,
		1221641927UL,
		1167319070UL,
		1095957929UL,
		1595256236UL,
		1540665371UL,
		1452775106UL,
		1381403509UL,
		1780037320UL,
		1859660671UL,
		1671105958UL,
		1733955601UL,
		2031960084UL,
		2111593891UL,
		1889500026UL,
		1952343757UL,
		2552477408UL,
		2632100695UL,
		2443283854UL,
		2506133561UL,
		2334638140UL,
		2414271883UL,
		2191915858UL,
		2254759653UL,
		3190512472UL,
		3135915759UL,
		3081330742UL,
		3009969537UL,
		2905550212UL,
		2850959411UL,
		2762807018UL,
		2691435357UL,
		3560074640UL,
		3505614887UL,
		3719321342UL,
		3648080713UL,
		3342211916UL,
		3287746299UL,
		3467911202UL,
		3396681109UL,
		4063920168UL,
		4143685023UL,
		4223187782UL,
		4286162673UL,
		3779000052UL,
		3858754371UL,
		3904687514UL,
		3967668269UL,
		881225847UL,
		809987520UL,
		1023691545UL,
		969234094UL,
		662832811UL,
		591600412UL,
		771767749UL,
		717299826UL,
		311336399UL,
		374308984UL,
		453813921UL,
		533576470UL,
		25881363UL,
		88864420UL,
		134795389UL,
		214552010UL,
		2023205639UL,
		2086057648UL,
		1897238633UL,
		1976864222UL,
		1804852699UL,
		1867694188UL,
		1645340341UL,
		1724971778UL,
		1587496639UL,
		1516133128UL,
		1461550545UL,
		1406951526UL,
		1302016099UL,
		1230646740UL,
		1142491917UL,
		1087903418UL,
		2896545431UL,
		2825181984UL,
		2770861561UL,
		2716262478UL,
		3215044683UL,
		3143675388UL,
		3055782693UL,
		3001194130UL,
		2326604591UL,
		2389456536UL,
		2200899649UL,
		2280525302UL,
		2578013683UL,
		2640855108UL,
		2418763421UL,
		2498394922UL,
		3769900519UL,
		3832873040UL,
		3912640137UL,
		3992402750UL,
		4088425275UL,
		4151408268UL,
		4197601365UL,
		4277358050UL,
		3334271071UL,
		3263032808UL,
		3476998961UL,
		3422541446UL,
		3585640067UL,
		3514407732UL,
		3694837229UL,
		3640369242UL,
		1762451694UL,
		1842216281UL,
		1619975040UL,
		1682949687UL,
		2047383090UL,
		2127137669UL,
		1938468188UL,
		2001449195UL,
		1325665622UL,
		1271206113UL,
		1183200824UL,
		1111960463UL,
		1543535498UL,
		1489069629UL,
		1434599652UL,
		1363369299UL,
		622672798UL,
		568075817UL,
		748617968UL,
		677256519UL,
		907627842UL,
		853037301UL,
		1067152940UL,
		995781531UL,
		51762726UL,
		131386257UL,
		177728840UL,
		240578815UL,
		269590778UL,
		349224269UL,
		429104020UL,
		491947555UL,
		4046411278UL,
		4126034873UL,
		4172115296UL,
		4234965207UL,
		3794477266UL,
		3874110821UL,
		3953728444UL,
		4016571915UL,
		3609705398UL,
		3555108353UL,
		3735388376UL,
		3664026991UL,
		3290680682UL,
		3236090077UL,
		3449943556UL,
		3378572211UL,
		3174993278UL,
		3120533705UL,
		3032266256UL,
		2961025959UL,
		2923101090UL,
		2868635157UL,
		2813903052UL,
		2742672763UL,
		2604032198UL,
		2683796849UL,
		2461293480UL,
		2524268063UL,
		2284983834UL,
		2364738477UL,
		2175806836UL,
		2238787779UL,
		1569362073UL,
		1498123566UL,
		1409854455UL,
		1355396672UL,
		1317987909UL,
		1246755826UL,
		1192025387UL,
		1137557660UL,
		2072149281UL,
		2135122070UL,
		1912620623UL,
		1992383480UL,
		1753615357UL,
		1816598090UL,
		1627664531UL,
		1707420964UL,
		295390185UL,
		358241886UL,
		404320391UL,
		483945776UL,
		43990325UL,
		106832002UL,
		186451547UL,
		266083308UL,
		932423249UL,
		861060070UL,
		1041341759UL,
		986742920UL,
		613929101UL,
		542559546UL,
		756411363UL,
		701822548UL,
		3316196985UL,
		3244833742UL,
		3425377559UL,
		3370778784UL,
		3601682597UL,
		3530312978UL,
		3744426955UL,
		3689838204UL,
		3819031489UL,
		3881883254UL,
		3928223919UL,
		4007849240UL,
		4037393693UL,
		4100235434UL,
		4180117107UL,
		4259748804UL,
		2310601993UL,
		2373574846UL,
		2151335527UL,
		2231098320UL,
		2596047829UL,
		2659030626UL,
		2470359227UL,
		2550115596UL,
		2947551409UL,
		2876312838UL,
		2788305887UL,
		2733848168UL,
		3165939309UL,
		3094707162UL,
		3040238851UL,
		2985771188UL
	};
	
#if MACHINE == CRAY
	for (crc = 0; *u != '\0'; u++)
		crc = ((crc << 8) & (~(~0UL << 32))) ^ crctab[(crc >> 24) ^ *u];
#else
	for (crc = 0; *u != '\0'; u++)
		crc = (crc << 8) ^ crctab[(crc >> 24) ^ *u];
#endif
	return(crc);
}

void handle_signal(int sig)
{
	static int nint = 0;
	
	if (sig == SIGINT)
	{
		nint++;
		if (nint < 2)
		{
			my_printf("caught SIGINT %d!\n", nint);
			BUG(signal(SIGINT, handle_signal) == SIG_ERR);
		}
		else
		{
			my_printf("SIGINT!\n");
			exit(EXIT_FAILURE);
		}
	}
}

#define START_BOARD "brd.txt"

/*
read_game reads the moves played sofar 
and returns the number of moves done by white and by black.
*/

local int read_game(FILE *f);
local int read_game(FILE *f)
{
	char line[LINE_MAX];
	char move[LINE_MAX];
	int iroot;
	int nwhite, nblack;
	
	nwhite = nblack = 0;
	while(my_fgets(line, LINE_MAX, f) != NULL)
	{
		if (sscanf(line, "%s", move) != 1) continue;
		gen_root();
		BUG((iroot = search_root(move, MOVE_STRING_ANY, FALSE, TRUE)) == INVALID)
		do_move(root[iroot].root_move);
		if (your_colour == WHITE_COLOUR)
			nwhite++;
		else
			nblack++;
	}
	compact_lists();
	if (my_colour == WHITE_COLOUR)
		return(nwhite);
	else
		return(nblack);
}

local void gen_book(char *);
local void gen_book(char *name)
{
	FILE *f, *g;
	char move[LINE_MAX];
	char string[LINE_MAX];
	int iroot;
	int imove, nmoves;
	
	fen_to_board(start_fen, TRUE);
	display_board();
	display_list();
	BUG((f = fopen(name, "r")) == NULL)
	BUG((g = fopen("dbase.txt", "w")) == NULL)
	nmoves = 0;
	while(fscanf(f, "%s", move) == 1)
	{
		if (strcmp(move, "!") == 0)
		{
			for (imove = 0; imove < nmoves; imove++)
				undo_move();
			nmoves = 0;
			continue;
		}
		gen_root();
		BUG((iroot = search_root(move, MOVE_STRING_ANY, FALSE, TRUE)) == INVALID)
		board_to_fen(string, FALSE);
		fprintf(g, "%s\n%s\n", string, move);
		do_move(root[iroot].root_move);
		nmoves++;
	}
	FCLOSE(f)
	FCLOSE(g)
}

/*
analyze_pos reads the positions from the file name, and analyzes them.
*/

local void analyze_pos(char *);
local void analyze_pos(char *name)
{
	FILE *f;
	int iroot;
	char string[LINE_MAX];
	
	BUG((f = fopen(name, "r")) == NULL)
	while(my_fgets(string, LINE_MAX, f) != NULL)
	{
		fen_to_board(string, FALSE);
		my_printf("pos board\n");
		display_board();
		display_list();
		display_score();
		display_attack();
		WHITE_TO_MOVE
		gen_root();
		my_printf("white moves:\n");
		for (iroot = 0; iroot < nroot; iroot++)
		{
			do_move(root[iroot].root_move);
			undo_move();
			move_to_string(root[iroot].root_move);
			my_printf(" %5s", move_string);
			move_to_san_string(root[iroot].root_move, TRUE);
			my_printf("(%7s)", move_string);
			if (((iroot % 5) == 4) or (iroot == (nroot - 1))) my_printf("\n");
		}
		BLACK_TO_MOVE
		gen_root();
		my_printf("black moves:\n");
		for (iroot = 0; iroot < nroot; iroot++)
		{
			do_move(root[iroot].root_move);
			undo_move();
			move_to_string(root[iroot].root_move);
			my_printf(" %5s", move_string);
			move_to_san_string(root[iroot].root_move, TRUE);
			my_printf("(%7s)", move_string);
			if (((iroot % 5) == 4) or (iroot == (nroot - 1))) my_printf("\n");
		}
	}
	FCLOSE(f)
}

/*
verify_game reads the move from the file name, and checks the moves
*/

local void verify_game(char *);
local void verify_game(char *name)
{
	FILE *f;
	char line[LINE_MAX];
	char move[LINE_MAX];
	int iroot;
	
	fen_to_board(start_fen, TRUE);	
	BUG((f = fopen(START_BOARD, "r")) == NULL)
	BUG(my_fgets(line, LINE_MAX, f) == NULL)
	FCLOSE(f)
	fen_to_board(line, TRUE);
	BUG((f = fopen(name, "r")) == NULL)
	while(my_fgets(line, LINE_MAX, f) != NULL)
	{
		if (sscanf(line, "%s", move) != 1) continue;
		gen_root();
		BUG((iroot = search_root(move, MOVE_STRING_ANY, FALSE, TRUE)) == INVALID)
		do_move(root[iroot].root_move);
	}
	FCLOSE(f)
}

/*
analyze_game reads the move from the file name, and analyzes the positions
before each move.
*/

local void analyze_game(char *);
local void analyze_game(char *name)
{
	FILE *f;
	char line[LINE_MAX];
	char move[LINE_MAX];
	int iroot;
	
	fen_to_board(start_fen, TRUE);
	BUG((f = fopen(START_BOARD, "r")) == NULL)
	BUG(my_fgets(line, LINE_MAX, f) == NULL)
	FCLOSE(f)
	fen_to_board(line, TRUE);
	display_board();
	display_score();
	
	BUG((f = fopen(name, "r")) == NULL)
	while(my_fgets(line, LINE_MAX, f) != NULL)
	{
		if (sscanf(line, "%s", move) != 1) continue;
		gen_root();
		BUG((iroot = search_root(move, MOVE_STRING_ANY, FALSE, TRUE)) == INVALID)
		display_score();
		do_root();
		do_move(root[iroot].root_move);
		display_board();
		display_score();
	}
	FCLOSE(f)
}

/*
solve_problems reads the problems from file name, and tries to solve them.
*/

local double solve_problems(char *);
local double solve_problems(char *name)
{
	FILE *f;
	char string[LINE_MAX];
	char move[LINE_MAX];
	int nproblems, nsolved, nsimple, ninvalid;
	int iroot;
	
	nproblems = nsolved = nsimple = ninvalid = 0;
	BUG((f = fopen(name, "r")) == NULL)
	while(my_fgets(string, LINE_MAX, f) != NULL)
	{
		fen_to_board(string, FALSE);
		BUG(my_fgets(string, LINE_MAX, f) == NULL)
		sscanf(string, "%s", move);
		nproblems++;
		display_board();
		display_list();
		
		gen_root();
		if (nroot == 0)
		{
			my_printf("problem %d is draw or mate!\n", nproblems);
			nsimple++;
			continue;
		}
		if (nroot == 1)
		{
			my_printf("problem %d has a forced move!\n", nproblems);
			nsimple++;
			continue;
		}
		iroot = search_root(move, MOVE_STRING_ANY, TRUE, TRUE);
		if (iroot == INVALID)
		{
			my_printf("problem %d has an invalid move!\n", nproblems);
			ninvalid++;
			continue;
		}
		do_root();
		if (iroot == jroot)
		{
			my_printf("solved problem %d!\n", nproblems);
			nsolved++;
		}
		else
			my_printf("did not solve problem %d!\n", nproblems);
	}
	my_printf("nproblems=%d nsolved=%d nsimple=%d ninvalid=%d\n",
		nproblems, nsolved, nsimple, ninvalid);
	FCLOSE(f)
	if (nproblems == 0)
		return((double) 0);
	else
		return((double) nsolved / (double) nproblems);
}

/*
find_mate reads the problems from file name, and tries to solve them.
*/

local void find_mate(char *);
local void find_mate(char *name)
{
	FILE *f;
	int nproblems, nsolved, nsimple;
	char string[LINE_MAX];
	
	nproblems = nsolved = nsimple = 0;
	BUG((f = fopen(name, "r")) == NULL)
	while(my_fgets(string, LINE_MAX, f) != NULL)
	{
		fen_to_board(string, FALSE);
		nproblems++;
		display_board();
		display_list();
		
		gen_root();
		if (nroot == 0)
		{
			my_printf("problem %d is draw or mate!\n", nproblems);
			nsimple++;
			continue;
		}
		if (nroot == 1)
		{
			my_printf("problem %d has a forced move!\n", nproblems);
			nsimple++;
			continue;
		}
		if (do_mate()) nsolved++;
	}
	my_printf("nproblems=%d nsolved=%d nsimple=%d\n",
		nproblems, nsolved, nsimple);
	FCLOSE(f)
}

/*
play_batch reads the moves played sofar from the file name, and appends 
a reply to the file name.
*/

local void play_batch(char *);
local void play_batch(char *name)
{
	FILE *f;
	char string[LINE_MAX];
	
	BUG((f = fopen(START_BOARD, "r")) == NULL)
	BUG(my_fgets(string, LINE_MAX, f) == NULL)
	FCLOSE(f)
	fen_to_board(string, TRUE);
	display_board();
	display_list();
	
	BUG((f = fopen(name, "r")) == NULL)
	read_game(f);
	FCLOSE(f)
	my_printf("board after read game\n");
	display_board();
	display_list();
	display_score();

	if (draw_by_repetition(TRUE))
	{
		my_printf("draw by rep!\n");
		exit(EXIT_FAILURE);
	}
	
	gen_root();
	if (nroot == 0)
	{
		if (i_am_check(my_colour, your_colour))
			my_printf("%sis mate!\n", colour_as_string[my_colour]);
		else
			my_printf("stale mate!\n");
		exit(EXIT_FAILURE);
	}
	if (nroot == 1)
	{
		my_printf("forced!\n");
		jroot = 0;
	}
	else
		do_root();
	if (Append)
	{
		BUG((f = fopen(name, "a")) == NULL)
		move_to_san_string(root[jroot].root_move, TRUE);
		fprintf(f, "%7s zzzzzz %.3f\n", move_string, (double) root[jroot].root_score);
		FCLOSE(f)
	}
}

#if MACHINE == MAC
#define GAME_TIME1 (10 * 60)
#define GAME_TIME2 (5 * 60)
#define NGAME_MOVES1 20
#define NGAME_MOVES2 40
#define GAME_MARGIN (1 * 60)
#else
#define GAME_TIME1 (90 * 60)
#define GAME_TIME2 (60 * 60)
#define NGAME_MOVES1 60
#define NGAME_MOVES2 40
#define GAME_MARGIN (5 * 60)
#endif

/*
play_game plays a game under tournament conditions.
zzzzzz plays xxxxxx, you play yyyyyy.
*/

local void play_game(int, int, char *);
local void play_game(int xxxxxx, int yyyyyy, char *name)
{
	FILE *f;
	int ngame_moves;
	time_t wall1, wall2;
	int wall;
	int iroot;
	char string[LINE_MAX];
	
	DUMMY("%d", yyyyyy)
	while(TRUE)
	{
		BUG((f = fopen(START_BOARD, "r")) == NULL)
		BUG(my_fgets(string, LINE_MAX, f) == NULL)
		FCLOSE(f)
		fen_to_board(string, TRUE);
		display_board();
		display_list();
		
		BUG((f = fopen(name, "r")) == NULL)
		ngame_moves = read_game(f);
		FCLOSE(f)
		my_printf("board after read game\n");
		display_board();
		/*
		display_list();
		display_score();
		*/
	
		if (draw_by_repetition(TRUE))
		{
			my_printf("draw by rep!\n");
			exit(EXIT_FAILURE);
		}
		
		gen_root();
		if (nroot == 0)
		{
			if (i_am_check(my_colour, your_colour))
				my_printf("%sis mate!\n", colour_as_string[my_colour]);
			else
				my_printf("stale mate!\n");
			exit(EXIT_FAILURE);
		}
		else
		{
			if (xxxxxx == my_colour)
			{
				if (nroot == 1)
				{
					my_printf("forced!\n");
					jroot = 0;
					root[jroot].root_score = 0;
				}
				else
				{
					/* calculate time limit */
					
					if (ngame_moves < NGAME_MOVES1)
					{
						TimeLimit =
						 	(GAME_TIME1 - GAME_MARGIN - GameTimeUsed) /
						 	(NGAME_MOVES1 - ngame_moves);
					}
					else if (ngame_moves < (NGAME_MOVES1 + NGAME_MOVES2))
					{
						TimeLimit =
							(GAME_TIME1 + GAME_TIME2 - GAME_MARGIN - GameTimeUsed) /
						  (NGAME_MOVES1 + NGAME_MOVES2 - ngame_moves);
					}
					else
					{
						TimeLimit = 90;
					}
					if (TimeLimit < 0)
					{
						my_printf("warning: negative time limit!\n");
						TimeLimit = 1;
					}
					my_printf("TimeLimit=%d\n", TimeLimit);
					wall1 = time(NULL);
					do_root();
					wall2 = time(NULL);
					wall = (int) difftime(wall2, wall1) + 1;
					GameTimeUsed += wall;
					my_printf("WallTime=%d GameTimeUsed=%d\n", wall, GameTimeUsed);
				}
				move_to_san_string(root[jroot].root_move, TRUE);
				my_printf("\azzzzzz move: %s\n", move_string);
				BUG((f = fopen(name, "a")) == NULL)
				fprintf(f, "%7s zzzzzz %.3f\n", move_string, (double) root[jroot].root_score);
				FCLOSE(f)
			}
			else
			{
				char line[LINE_MAX], move[LINE_MAX];
				
				if ((ngame_moves > 0) and (ngame_moves % 5) == 0)
				{
					while(TRUE)
					{
						my_printf("\aGameTimeUsed? (min) ");
						if (gets(line) == NULL) exit(EXIT_FAILURE);
						if (sscanf(line, "%d", &GameTimeUsed) != 1) continue;
						GameTimeUsed *= 60;
						break;
					}
				}
				/*
				Just for information.
				*/
				
				get_book_move();
				while(TRUE)
				{
					my_printf("your move? ");
					if (gets(line) == NULL) exit(EXIT_FAILURE);
					if (sscanf(line, "%s", move) != 1) continue;
					if (strcmp(move, "quit") == 0) exit(EXIT_SUCCESS);
					if (strcmp(move, "?") == 0)
					{
						for (iroot = 0; iroot < nroot; iroot++)
						{
							move_to_san_string(root[iroot].root_move, FALSE);
							my_printf(" %7s", move_string);
							if (((iroot % 9) == 8) or (iroot == (nroot - 1))) my_printf("\n");
						}
						continue;
					}
					iroot = search_root(move, MOVE_STRING_ANY, FALSE, FALSE);
					if (iroot == INVALID) continue;
					BUG((f = fopen(name, "a")) == NULL)
					move_to_san_string(root[iroot].root_move, TRUE);
					fprintf(f, "%7s\n", move_string);
					FCLOSE(f)
					break;
				}
			}
		}
	}
}

#define NDIM 23
#define PROBLEMS_FILE "opt.txt"
#define INIT_FILE "init.txt"
#define RESULTS_FILE "results.txt"

local void try(void);
local void try(void)
{
	FILE *f, *g;
	double err;
	int i;
	double y[NDIM];
	
	BUG((f = fopen(INIT_FILE, "r")) == NULL)
	BUG((g = fopen(RESULTS_FILE, "a")) == NULL)
	while(fscanf(f, "%lf", &err) == 1)
	{
		BUG(err > 1.0)
		for (i = 0; i < NDIM; i++)
			BUG(fscanf(f, "%lf", y + i) != 1)
		if (err < 0)
		{
			set_score(y);
			err = solve_problems(PROBLEMS_FILE);
			fprintf(g, "%.2f", err);
			for (i = 0; i < NDIM; i++)
				fprintf(g, " %d", (int) y[i]);
			fprintf(g, "\n");
		}
	}
	FCLOSE(g)
	FCLOSE(f)
}

options_t options;

local void set_options(void);
local void set_options(void)
{
	options.random_opening = 0;
	options.print_dbase = 0;
	options.print_distrib = 0;
}

local void read_config(char *);
local void read_config(char *name)
{
	FILE *f;
	char line[LINE_MAX];

	f = fopen(name, "r");
	if (f == NULL) return;
	my_printf("reading config from %s\n", name);
	while(my_fgets(line, LINE_MAX, f) != NULL)
	{
		if (*line == '#') continue;
		if (sscanf(line, "DepthMax=%d", &DepthMax) == 1)
		{
			my_printf("DepthMax=%d\n", DepthMax);
		}
		else if (sscanf(line, "TimeLimit=%d", &TimeLimit) == 1)
		{
			my_printf("TimeLimit=%d\n", TimeLimit);
		}
		else if (sscanf(line, "random_opening=%d", &(options.random_opening)) == 1)
		{
			my_printf("random_opening=%d\n", options.random_opening);
		}
		else if (sscanf(line, "print_dbase=%d", &(options.print_dbase)) == 1)
		{
			my_printf("print_dbase=%d\n", options.print_dbase);
		}
		else if (sscanf(line, "print_distrib=%d", &(options.print_distrib)) == 1)
		{
			my_printf("print_distrib=%d\n", options.print_distrib);
		}
		else
		{
			my_printf("unknown config:%s\n", line);
		}
	}
}

local void read_env(void);
local void read_env(void)
{
	char *p;
	
	my_printf("reading config from environment\n");
	p = getenv("ZZZZZZ_TIMELIMIT");
	if (p != NULL)
	{
		TimeLimit = atoi(p);
		my_printf("TimeLimit=%d\n", TimeLimit);
	}
	p = getenv("ZZZZZZ_DEPTHMAX");
	if (p != NULL)
	{
		DepthMax = atoi(p);
		my_printf("DepthMax=%d\n", DepthMax);
	}
}

#if MACHINE == MAC
#define ARGC_MAX 256
int argc;
char *argv_table[ARGC_MAX];
char **argv = argv_table;
int main(void);
int main(void)
#else
int main(int argc, char **argv);
int main(int argc, char **argv)
#endif
{
	char version_string[VERSION_STRING_MAX];
	char *p;
	char config[FILENAME_MAX];
	
#if MACHINE == MAC
	{
		FILE *f;
		char line[LINE_MAX];
	
		BUG((f = fopen("arg.txt", "r")) == NULL)
		argc = 0;
		while (fscanf(f, "%s", line) == 1)
		{
			BUG(argc >= ARGC_MAX)
			MALLOC(argv_table[argc], char, strlen(line) + 1)
			strcpy(argv_table[argc], line);
			argc++;
		}
		FCLOSE(f)
	}
#else
	argc--;
	argv++;
#endif

	set_options();

	zzzzzz_version(version_string);
	my_printf("zzzzzz version %s\n", version_string);
	util_version(version_string);
	my_printf("util   version %s\n", version_string);
	pgn_version(version_string);
	my_printf("pgn    version %s\n", version_string);
	
	signal(SIGINT, handle_signal);
	
	my_printf("ZZZZZZ version 3.4, Copyright (C) 1993-1994 G. Wiesenekker\n");
	my_printf("ZZZZZZ comes with ABSOLUTELY NO WARRANTY.\n");
	my_printf("ZZZZZZ is free software, and you are welcome to redistribute it\n");
	my_printf("under certain conditions.\n");
	my_printf("See the file COPYING (of which you should have received a copy\n");
	my_printf("along with ZZZZZZ) for details.\n");

	/* parse command line options */
	
	if ((p = getenv("HOME")) != NULL)
	{
		sprintf(config, "%s/.zzzzzzrc", p);
		read_config(config);
	}
	read_config(".zzzzzzrc");
	read_env();
	
	my_printf("reading command line arguments\n");
	while((argc > 0) and (*argv[0] == '-'))
	{
		if (strcmp(argv[0], "-DepthMax") == 0)
		{
			BUG(argc < 2)
			BUG(sscanf(argv[1], "%d", &DepthMax) != 1)
			argc--; argc--;
			argv++; argv++;
		}
		else if (strcmp(argv[0], "-TimeLimit") == 0)
		{
			BUG(argc < 2)
			BUG(sscanf(argv[1], "%d", &TimeLimit) != 1)
			argc--; argc--;
			argv++; argv++;
		}
		else if (strcmp(argv[0], "-GameTimeUsed") == 0)
		{
			BUG(argc < 2)
			BUG(sscanf(argv[1], "%d", &GameTimeUsed) != 1)
			BUG(GameTimeUsed < 0)
			argc--; argc--;
			argv++; argv++;
		}
		else if (strcmp(argv[0], "-MateDepth") == 0)
		{
			BUG(argc < 2)
			BUG(sscanf(argv[1], "%d", &MateDepth) != 1)
			BUG(MateDepth < 0)
			argc--; argc--;
			argv++; argv++;
		}
		else if (strcmp(argv[0], "-NoAppend") == 0)
		{
			Append = FALSE;
			argc--;
			argv++;
		}
		else if (strcmp(argv[0], "-NoBook") == 0)
		{
			UseBook = FALSE;
			argc--;
			argv++;
		}
		else if (strcmp(argv[0], "-NoPos") == 0)
		{
			UsePos = FALSE;
			argc--;
			argv++;
		}
#ifdef PVM
		else if (strcmp(argv[0], "-PvmProcs") == 0)
		{
			BUG(argc < 2)
			BUG(sscanf(argv[1], "%d", &PvmProcs) != 1)
			BUG(PvmProcs < 1)
			BUG(PvmProcs >= PVM_PROCS_MAX)
			argc--; argc--;
			argv++; argv++;
			do_master();
		}
		else if (strcmp(argv[0], "-PvmDepth") == 0)
		{
			BUG(argc < 2)
			BUG(sscanf(argv[1], "%d", &PvmDepth) != 1)
			BUG(PvmDepth < 0)
			argc--; argc--;
			argv++; argv++;
		}
		else if (strcmp(argv[0], "-PvmLegal") == 0)
		{
			BUG(argc < 2)
			BUG(sscanf(argv[1], "%d", &PvmLegal) != 1)
			BUG(PvmLegal < 0)
			argc--; argc--;
			argv++; argv++;
		}
#endif
		else
		{
			my_printf("unknown option %s ignored\n", argv[0]);
			argc--;
			argv++;
		}
	}
	
	my_printf("DepthMax=%d TimeLimit=%d GameTimeUsed=%d MateDepth=%d\n",
		DepthMax, TimeLimit, GameTimeUsed, MateDepth);
	my_printf("UseBook=%d UsePos=%d\n",
		UseBook, UsePos);
#ifdef PVM
	my_printf("PvmProcs=%d PvmDepth=%d PvmLegal=%d\n",
		PvmProcs, PvmDepth, PvmLegal);
#endif
	GameTimeUsed *= 60;
	
	BUG(argc < 1)
	initialize();

	if (strcmp(argv[0], "pos") == 0)
	{
		BUG(argc < 2)
		analyze_pos(argv[1]);
	}
	else if (strcmp(argv[0], "verif") == 0)
	{
		BUG(argc < 2)
		verify_game(argv[1]);
	}
	else if (strcmp(argv[0], "anal") == 0)
	{
		BUG(argc < 2)
		analyze_game(argv[1]);
	}
	else if (strcmp(argv[0], "solve") == 0)
	{
		BUG(argc < 2)
		solve_problems(argv[1]);
	}
	else if (strcmp(argv[0], "mate") == 0)
	{
		BUG(argc < 2)
		find_mate(argv[1]);
	}
	else if (strcmp(argv[0], "batch") == 0)
	{
		BUG(argc < 2)
		play_batch(argv[1]);
	}
	else if (strcmp(argv[0], "white") == 0)
	{
		BUG(argc < 2)
		play_game(WHITE_COLOUR, BLACK_COLOUR, argv[1]);
	}
	else if (strcmp(argv[0], "black") == 0)
	{
		BUG(argc < 2)
		play_game(BLACK_COLOUR, WHITE_COLOUR, argv[1]);
	}
	else if (strcmp(argv[0], "try") == 0)
	{
		try();
	}
	else if (strcmp(argv[0], "book") == 0)
	{
		BUG(argc < 2)
		gen_book(argv[1]);
	}
	else if (strcmp(argv[0], "pgn") == 0)
	{
		pgn();
	}
	else if (strcmp(argv[0], "crc") == 0)
	{
		gen_crc();
	}
	else if (strcmp(argv[0], "dbase") == 0)
	{
		BUG(argc < 2)
		dbase(argv[1]);
	}
	else if (strcmp(argv[0], "btree") == 0)
	{
		btree();
	}
#ifdef PVM
	else if (strcmp(argv[0], "driver") == 0)
	{
		do_driver();
	}
	else if (strcmp(argv[0], "slave") == 0)
	{
		do_slave();
	}
#endif
	else
	{
		my_printf("unknown arg %s!\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	return(EXIT_SUCCESS);
}
