/*
	G. Wiesenekker. ZZZZZZ a chess program.
	Copyright (C) 1993-1994 G. Wiesenekker
	E-mail: wiesenekker@sara.nl

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "zzzzzz.h"

/*
this saves some memory (the __FILE__ string is only generated once).
*/

local char *FILE_NAME = __FILE__;

#ifndef DEBUG_LEVEL
	#define DEBUG_LEVEL 0
#endif

void zzzzzz_version(char *version_string)
{
	sprintf(version_string, "%s %s %d", __DATE__, __TIME__, DEBUG_LEVEL);
}

/*
dummy is always false, but the compiler does not know that, so DUMMY
can be used to disable some compiler warnings.
*/

int dummy = FALSE;

int DepthMax = 32;
int TimeLimit = 30;
int GameTimeUsed = 0;
int MateDepth = 6;
int Append = 1;
int UseBook = 1;
int UsePos = 1;
int ZapCache = 0;
#ifdef PVM
int PvmProcs = 2;
int PvmDepth = 2;
int PvmLegal = 2;
#endif

/*
my_colour and your_colour are toggled by do_move and undo_move
*/

int my_colour, your_colour;

/*
I know about cylinder boards, but they give me a headache!
I use a 12x12 board, (8x8 with a double boundary).
*/

#define BOARD_MAX 144

enum {A8 =      26, B8, C8, D8, E8, F8, G8, H8};
enum {A7 = A8 + 12, B7, C7, D7, E7, F7, G7, H7};
enum {A6 = A7 + 12, B6, C6, D6, E6, F6, G6, H6};
enum {A5 = A6 + 12, B5, C5, D5, E5, F5, G5, H5};
enum {A4 = A5 + 12, B4, C4, D4, E4, F4, G4, H4};
enum {A3 = A4 + 12, B3, C3, D3, E3, F3, G3, H3};
enum {A2 = A3 + 12, B2, C2, D2, E2, F2, G2, H2};
enum {A1 = A2 + 12, B1, C1, D1, E1, F1, G1, H1};

enum {a8 =      0, b8, c8, d8, e8, f8, g8, h8};
enum {a7 = a8 + 8, b7, c7, d7, e7, f7, g7, h7};
enum {a6 = a7 + 8, b6, c6, d6, e6, f6, g6, h6};
enum {a5 = a6 + 8, b5, c5, d5, e5, f5, g5, h5};
enum {a4 = a5 + 8, b4, c4, d4, e4, f4, g4, h4};
enum {a3 = a4 + 8, b3, c3, d3, e3, f3, g3, h3};
enum {a2 = a3 + 8, b2, c2, d2, e2, f2, g2, h2};
enum {a1 = a2 + 8, b1, c1, d1, e1, f1, g1, h1};

/*
Sometimes we have to convert from the 'normal' 
index (0..63) to the 'true' index.
*/

local int map[64] =
{
	 A8, B8, C8, D8, E8, F8, G8, H8,
	 A7, B7, C7, D7, E7, F7, G7, H7,
	 A6, B6, C6, D6, E6, F6, G6, H6,
	 A5, B5, C5, D5, E5, F5, G5, H5,
	 A4, B4, C4, D4, E4, F4, G4, H4,
	 A3, B3, C3, D3, E3, F3, G3, H3,
	 A2, B2, C2, D2, E2, F2, G2, H2,
	 A1, B1, C1, D1, E1, F1, G1, H1
};

/*
mirror_map comes in handy when we have to initialize
tables which are different for white and black.
*/

local int mirror_map[64] =
{
	 A1, B1, C1, D1, E1, F1, G1, H1,
	 A2, B2, C2, D2, E2, F2, G2, H2,
	 A3, B3, C3, D3, E3, F3, G3, H3,
	 A4, B4, C4, D4, E4, F4, G4, H4,
	 A5, B5, C5, D5, E5, F5, G5, H5,
	 A6, B6, C6, D6, E6, F6, G6, H6,
	 A7, B7, C7, D7, E7, F7, G7, H7,
	 A8, B8, C8, D8, E8, F8, G8, H8
};

/* 
This needs no comment
*/

local char *nota_table[64] = 
{
	"a8", "b8", "c8", "d8", "e8", "f8", "g8", "h8",
	"a7", "b7", "c7", "d7", "e7", "f7", "g7", "h7",
	"a6", "b6", "c6", "d6", "e6", "f6", "g6", "h6",
	"a5", "b5", "c5", "d5", "e5", "f5", "g5", "h5",
	"a4", "b4", "c4", "d4", "e4", "f4", "g4", "h4",
	"a3", "b3", "c3", "d3", "e3", "f3", "g3", "h3",
	"a2", "b2", "c2", "d2", "e2", "f2", "g2", "h2",
	"a1", "b1", "c1", "d1", "e1", "f1", "g1", "h1"
};
local char *nota_inval = "??";
local char **nota;

#define GET_NOTA(X) ((((X) < 0) or ((X) > BOARD_MAX)) ? "??" : nota[X])

/*
The row numbers used by the program. 
*/

local byte_t row_table[64] =
{
	7, 7, 7, 7, 7, 7, 7, 7,
	6, 6, 6, 6, 6, 6, 6, 6,
	5, 5, 5, 5, 5, 5, 5, 5,
	4, 4, 4, 4, 4, 4, 4, 4,
	3, 3, 3, 3, 3, 3, 3, 3,
	2, 2, 2, 2, 2, 2, 2, 2,
	1, 1, 1, 1, 1, 1, 1, 1,
	0, 0, 0, 0, 0, 0, 0, 0
};
local int *row;
local char *nota_row = "12345678";
#define GET_ROW(X) nota_row[row[X]]

/*
The column numbers used by the program
*/

local byte_t column_table[64] =
{
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7,
	0, 1, 2, 3, 4, 5, 6, 7
};
local int *column;
local char *nota_column = "abcdefgh";
#define GET_COL(X) nota_column[column[X]]

local char *nota_colour = "wb.x";
char *colour_as_string[2] = {"white ", "black "};

enum {PAWN = 0, KNIGHT, BISHOP, ROOK, QUEEN, KING};

local char *nota_kind = "PNBRQK";

/*
To detect double pawns, free pawns, strong pieces etc. we need
information about the pawn structure. We store the pawn structure
in 8 bytes (one for each column) and if a square on a column
is occupied by a pawn, the corresponding row bit in the byte for 
that column is set. The following tables are initialized at startup
bit_count: gives the number of bits set in a byte.
bit_low: gives the lowest bit set in a byte.
bit_high: gives the highest bit set in a byte.
*/

#define NBITS 8
#define NBITS_SIZE BIT(NBITS)

local int *bit_count;
local int *bit_low;
local int *bit_high;
local int white_occup[8], black_occup[8];
local int *occup[2] = {white_occup, black_occup};

/*
Some more evaluation function macro's. SCORE_LIMIT stands for infinity.
If an endgame is detected which is won, SCORE_ENDGAME is added to the score
to make sure the program makes a transition to the endgame.
Of course, in some positions this leads to strange play: the program 
sacrifices a pawn instead of promoting it to a queen in order to make the 
transition to the endgame.
If the score is greater than SCORE_ABORT, we may abort the search because
the program found mate.
*/

#define SCORE_LIMIT   30000
#define SCORE_MATE    28000
#define SCORE_ABORT   26000
#define SCORE_ENDGAME 20000

/*
The evaluation function parameters. They just run from 'bad'
to 'good' and increase by 5 or 10 points. Nevertheless 
the program scored 5 out of 11 in the Dutch Computerchess Championships 1993.
*/

local int weak_queen_score = -270;
local int weak_rook_score = -220;
local int weak_bishop_score = -170;
local int weak_knight_score = -120;
local int isolated_pawn_score = -70;
local int king_danger_score = -40;
local int double_pawn_score = -30;
local int lagged_pawn_score = -20;
local int bishop_blocks_pawn_score = -15;
local int bishop_not_developed_score = -5;
local int knight_not_developed_score = -5;
local int weak_pawn_attack_score = 10;
local int pin_score = 20;
local int rook_on_half_open_file_score = 25;
local int knight_strong_score = 30;
local int bishop_strong_score = 35;
local int rook_on_open_file_score = 40;
local int rook_on_7th_rank_score = 50;
local int long_castle_score = 60;
local int short_castle_score = 70;
local int knight_can_move_score = 2;
local int bishop_can_move_score = 3;
local int rook_can_move_score = 1;

local int value[6] = 
{
	PAWN_SCORE, KNIGHT_SCORE, BISHOP_SCORE, ROOK_SCORE, QUEEN_SCORE, KING_SCORE
};

/* 
The king gradient is used in simple endgames. It forces the king to the
side of the board.
*/

local byte_t king_alone_grad_table[64] =
{
	-64,-32,-16,-16,-16,-16,-32,-64,
	-32,-16, -8, -8, -8, -8,-16,-32,
	-16, -8, -4, -4, -4, -4, -8,-16,
	-16, -8, -4, -2, -2, -4, -8,-16,
	-16, -8, -4, -2, -2, -4, -8,-16,
	-16, -8, -4, -4, -4, -4, -8,-16,
	-32,-16, -8, -8, -8, -8,-16,-32,
	-64,-32,-16,-16,-16,-16,-32,-64
};
local int *king_alone_grad;

/*
The pawn gradient urges pawns to promote.
*/

local byte_t free_pawn_grad_table[64] =
{
	 0,  0,  0,  0,  0,  0,  0,  0,
	64, 96, 96, 96, 96, 96, 96, 64,
	32, 48, 48, 48, 48, 48, 48, 32,
	16, 24, 24, 24, 24, 24, 24, 16,
	 8, 12, 12, 12, 12, 12, 12,  8,
	 4,  6,  6,  6,  6,  6,  6,  4,
	 2,  3,  3,  3,  3,  3,  3,  2,
	 0,  0,  0,  0,  0,  0,  0,  0
};
local int *white_free_pawn_grad, *black_free_pawn_grad;
local int *free_pawn_grad[2];

/*
Visit tables for pawn and knight (only some fields are used in this version).
*/

local byte_t pawn_visit_table[64] = 
{
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  3,  3,  0,  0,  0,
	 0,  0,  1,  2,  2,  1,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0
};

local byte_t knight_visit_table[64] = 
{
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0,
	 0,  0,  3,  0,  0,  3,  0,  0,
	 0,  0,  0,  2,  2,  0,  0,  0,
	 0,  0,  0,  0,  0,  0,  0,  0
};

local int *white_pawn_visit, *black_pawn_visit;
local int *pawn_visit[2];

local int *white_knight_visit, *black_knight_visit;
local int *knight_visit[2];

local int *pawn_board;

/*
rooks on 7-th rank are strong.
*/

local int strong_row[2] = {6, 1};

/*
attack_zero is used to clear the attack array.
*/

local int attack_zero[BOARD_MAX] = 
{
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1, 0, 0, 0, 0, 0, 0, 0, 0,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,
	-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1
};

local int *attack_board;

/*
movegen stuff
*/

#if MACHINE == DOS
#define MOVE_MAX 128
#define CAPT_MAX 1024
#else
#define MOVE_MAX 256
#define CAPT_MAX 2048
#endif

/*
We pack the move (from, to and kind) in a long.
ADD_MOVE pushes a move on a stack. P points to the top of the stack.
*/

#define PACK_MOVE(X, K, A, B)\
	(X) = ((K) | ((B) << 8) | (A));
#define UNPACK_MOVE(X, K, A, B)\
	(K) = ((X) & 0xFFFF0000L);\
	(A) = (int) ((X) & 0x000000FFL);\
	(B) = (int) (((X) >> 8) & 0x000000FFL);
#define ADD_MOVE(P, K, A, B)\
	{\
		PACK_MOVE(*(P), K, A, B)\
		(P)++;\
	}

/*
We generate moves in a smart order: pawn captures first, then
knight, bishop, rook, queen and king captures. Then en-passant and castling moves
and finally non-capturing moves. When we search for captures, we store the
captures and the special moves in a capture stack, and the normal moves
in a move stack. move_top and capt_top are pointers to the current top of the
move and capture stacks.
*/

local move_t *move_stack, *capt_stack;
local move_t *move_top;
move_t *capt_top;

local int shift[6] = 
	{FALSE, FALSE, TRUE, TRUE, TRUE, FALSE};

/*
The offsets to generate the moves. We use the forward directions first.
This may give more cutoffs (the moves are generated in 
a more natural order) but I never checked how much.
The end of the list is denoted by offset 0. So you will often see
loops like

board_t *p;
int *d;
for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
{
	if (p->board_where == INVALID) continue;
	for (d = p->board_dir; *d != 0; d++)
	{
		..
	}
}

*/

local int white_pawn_dir[4] =
	{-12,   -13, -11, 0};
local int black_pawn_dir[4] =
	{ 12,    11,  13, 0};
local int *pawn_dir[2] =
	{white_pawn_dir, black_pawn_dir};

local int white_knight_dir[9] =
	{-25, -23, -14, -10,  10, 14,  23,   25,  0};
local int black_knight_dir[9] =
	{ 25,  23,  14,  10, -10, -14, -23, -25,  0};
local int *knight_dir[2] =
	{white_knight_dir, black_knight_dir};

local int white_bishop_dir[5] =
	{-13, -11,  11,  13, 0};
local int black_bishop_dir[5] =
	{ 13,  11, -11, -13, 0};
local int *bishop_dir[2] =
	{white_bishop_dir, black_bishop_dir};

local int white_rook_dir[5] =
	{-12,  -1,   1,  12, 0};
local int black_rook_dir[5] =
	{ 12,   1,  -1, -12, 0};
local int *rook_dir[2] =
	{white_rook_dir, black_rook_dir};

local int white_queen_dir[9] =
	{-13, -12, -11, -1,   1,  11,  12,  13,  0};
local int black_queen_dir[9] =
	{ 13,  12,  11,  1,  -1, -11, -12, -13,  0};

local int white_king_dir[9] = 
	{-13, -12, -11,  -1,  1,  11,  12,  13,  0};
local int black_king_dir[9] = 
	{ 13,  12,  11,   1, -1, -11, -12, -13,  0};

local int *white_dir[6] =
	{white_pawn_dir, white_knight_dir, white_bishop_dir, white_rook_dir, white_queen_dir, white_king_dir};
local int *black_dir[6] =
	{black_pawn_dir, black_knight_dir, black_bishop_dir, black_rook_dir, black_queen_dir, black_king_dir};
local int **dir[2] =
	{white_dir, black_dir};
	
/*
A pawn promotes if it is on row 7 for white or row 0 for black.
*/
	
local int prom_row[2] = {7, 0};

/*
hack_ep keeps track of possible en passant right (will be INVALID
most of the time.
*/

local int hack_ep;

/*
We store the pieces in lists, one list for every piece.
The board is an array of pointers to these lists. Two special
list entries are the empty square and the side square.
*/

local board_t *board[BOARD_MAX];
local board_t empty = {NONE, INVALID, 0, INVALID, TRUE, FALSE, NULL, FALSE, NULL};
local board_t side = {SIDE, INVALID, 0, INVALID, TRUE, FALSE, NULL, FALSE, NULL};

local board_t white_pawn_list[LIST_MAX], black_pawn_list[LIST_MAX];
local board_t *pawn_list[2] =
	{white_pawn_list, black_pawn_list};

local board_t white_knight_list[LIST_MAX], black_knight_list[LIST_MAX];
local board_t *knight_list[2] =
	{white_knight_list, black_knight_list};

local board_t white_bishop_list[LIST_MAX], black_bishop_list[LIST_MAX];
local board_t *bishop_list[2] =
	{white_bishop_list, black_bishop_list};

local board_t white_rook_list[LIST_MAX], black_rook_list[LIST_MAX];
local board_t *rook_list[2] =
	{white_rook_list, black_rook_list};

local board_t white_queen_list[LIST_MAX], black_queen_list[LIST_MAX];
local board_t *queen_list[2] =
	{white_queen_list, black_queen_list};

local board_t white_king_list[LIST_MAX], black_king_list[LIST_MAX];
local board_t *king_list[2] =
	{white_king_list, black_king_list};

local board_t *white_list[6] =
	{white_pawn_list, white_knight_list, white_bishop_list, white_rook_list, white_queen_list, white_king_list};
local board_t *black_list[6] =
	{black_pawn_list, black_knight_list, black_bishop_list, black_rook_list, black_queen_list, black_king_list};
local board_t **list[2] =
	{white_list, black_list};
	
local int white_piece_count[6];
local int black_piece_count[6];
local int	*piece_count[2] = {white_piece_count, black_piece_count};
int tot_count[2];

/*
VAR_MAX is the maximum length of the principle variation.
*/

#define VAR_MAX 256

local var_t *var_stack;

#if MACHINE == DOS
#define NODE_MAX 256
#else
#define NODE_MAX 1024
#endif

local node_t *node_stack;
node_t *node;
local int inode, inode_root, inode_max;

/* 
NULL_MOVE_DEPTH is the minimum search depth at which we may
do a null move. We search regular moves at (depth - 1), 
but null moves are searched at (depth - 1 - NULL_MOVE_REDUC).
If the null move score fails high, the node is pruned.
(This fails terribly in case of zugzwang).
*/

#define NULL_MOVE_DEPTH 3
#define NULL_MOVE_REDUC 2

/*
PAR_DEPTH is the minimum search depth at which we may 
distribute the moves over a number of slaves.
PAR_DEPTH should be large enough so that the time needed
for the search is larger than the time needed for the
distribution of the moves.
*/

#define PAR_DEPTH 3

/*
The program uses two hash tables: one for the alpha-beta, and one for 
the static evaluation.
The hash table sizes are powers of two, so we can calculate the index
by (hask_key.low & (~(~0L << CACHE_BITS))).
Since we then already know the value of the lower CACHE_BITS bits, 
we can use these bits to store other useful information.
*/
#undef LARGE
#ifdef SMALL
	#define CACHE_BITS        9
	#define SCORE_CACHE_BITS  9
#else
	#if MACHINE == DOS
		#define CACHE_BITS        12
		#define SCORE_CACHE_BITS  12
	#elif MACHINE == MAC
		#ifdef LARGE
			#define CACHE_BITS        14
			#define SCORE_CACHE_BITS  15
		#else
			#define CACHE_BITS        13
			#define SCORE_CACHE_BITS  14
		#endif
	#else
		#define CACHE_BITS        16
		#define SCORE_CACHE_BITS  17
	#endif
#endif

#define CACHE_MASK        (~(~0L << CACHE_BITS))
#define SCORE_CACHE_MASK  (~(~0L << SCORE_CACHE_BITS))

/*
We use CACHE_DEPTH_BITS bits for the depth, so the maximum search depth than can
be stored in the hash table is CACHE_DEPTH_MASK.
*/

#ifdef SMALL
#define CACHE_DEPTH_BITS 4
#else
#define CACHE_DEPTH_BITS 6
#endif
#define CACHE_DEPTH_MASK (~(~0UL << CACHE_DEPTH_BITS))

/*
CACHE_INVAL is set if the hash table entry is empty.
*/

#define CACHE_INVAL BITL(CACHE_DEPTH_BITS)
#define CACHE_LOWER BITL(CACHE_DEPTH_BITS + 1)
#define CACHE_UPPER BITL(CACHE_DEPTH_BITS + 2)
#define CACHE_TRUE  BITL(CACHE_DEPTH_BITS + 3)

#define CACHE_KIND_BITS (CACHE_DEPTH_BITS + 4)
#define CACHE_KIND_MASK (~(~0L << CACHE_KIND_BITS))

#if CACHE_BITS < CACHE_KIND_BITS
#error CACHE_BITS should not be smaller than CACHE_KIND_BITS
#endif

/*
the pointers to the tables with hash keys.
*/

#if MACHINE == DOS
local hash_key_t huge *hash_key[2][6];
local hash_key_t huge *ep_key;
local hash_key_t huge *castle_key;
#else
local hash_key_t *hash_key[2][6];
local hash_key_t *ep_key;
local hash_key_t *castle_key;
#endif

/*
cache_best is a pointer to the best move (it is the number of the best move
in the list of pseudo legal moves).
cache_score is the score.
cache_key is the key
*/

#define FEN_STRING_MAX 96

typedef struct
{
	int cache_best;
	score_t cache_score;
	hash_key_t cache_key;
#ifdef CACHE_FEN
	char cache_fen[FEN_STRING_MAX];
#endif
} cache_t;

/*
For the static evaluation hash table we do not have a best move.
*/

typedef struct
{
	score_t cache_score;
	hash_key_t cache_key;
#ifdef CACHE_FEN
	char cache_fen[FEN_STRING_MAX];
#endif
} score_cache_t;

/*
We use separate hash tables for white and black
*/

#if MACHINE == DOS
local cache_t huge *cache[2];
local score_cache_t huge *score_cache[2];
#else
local cache_t *cache[2];
local score_cache_t *score_cache[2];
#endif

/*
Some counters for statistics.
*/

local long ngens, ndos;
local long nevals, nhits;
#ifdef BUG_DEPTH
local int bug_depth[NODE_MAX];
#endif

/*
Variables for time control
*/

local clock_t clock_begin, clock_end;
local double time_used;

/*
move_to_string converts the move to a long algebraic notation 
and stores the move in move_string.
*/

char move_string[MOVE_STRING_MAX];

void move_to_string(move_t m)
{
	long kind;
	int from, to;
	char c;
	
	UNPACK_MOVE(m, kind, from, to)
	if (kind == NULL_MOVE)
	{
		sprintf(move_string, "%s", "NUL");
	}
	else if (kind & PROM_CAPT)
	{
		if (kind & PROM_QUEEN)
			c = 'Q';
		else if (kind & PROM_ROOK)
			c = 'R';
		else if (kind & PROM_BISHOP)
			c = 'B';
		else
			c = 'N';
		sprintf(move_string, "%s%s%c", GET_NOTA(from), GET_NOTA(to), c);
	}
	else if (kind & PROM_NORM)
	{
		if (kind & PROM_QUEEN)
			c = 'Q';
		else if (kind & PROM_ROOK)
			c = 'R';
		else if (kind & PROM_BISHOP)
			c = 'B';
		else
			c = 'N';
		sprintf(move_string, "%s%s%c", GET_NOTA(from), GET_NOTA(to), c);
	}
	else if (kind == EP)
		sprintf(move_string, "%s%s", GET_NOTA(from), GET_NOTA(to));
	else if (kind == SHORT)
		sprintf(move_string, "o-o");
	else if (kind == LONG)
		sprintf(move_string, "o-o-o");
	else if (kind == CAPT)
		sprintf(move_string, "%s%s", GET_NOTA(from), GET_NOTA(to));
	else
		sprintf(move_string, "%s%s", GET_NOTA(from), GET_NOTA(to));
}

void display_board(void)
{
	int i, j, k;
	
	k = 0;
	for (j = 0; j < 8; j++)
	{
		for (i = 0; i < 8; i++)
		{
			if (board[map[k]] == &empty)
				my_printf("..");
			else
				my_printf("%c%c",
					nota_colour[board[map[k]]->board_colour],
					nota_kind[board[map[k]]->board_kind]);
			k++;
		}
		my_printf("\n");
	}
	my_printf("%s\n", colour_as_string[my_colour]);
}

/*
board_to_fen converts a board position to FEN notation
*/

local char *white_fen[6] = {"P", "N", "B", "R", "Q", "K"};
local char *black_fen[6] = {"p", "n", "b", "r", "q", "k"};
local char *empty_fen[9] = {"", "1", "2", "3", "4", "5", "6", "7", "8"};
local char *file_fen[9] = {"a", "b", "c", "d", "e", "f", "g", "h"};
char *start_fen = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1";
local char *start_board = "rnbqkbnr/pppppppp/eeeeeeee/eeeeeeee/eeeeeeee/eeeeeeee/PPPPPPPP/RNBQKBNR";

/*
We also use the fen string for the database. As we are then
only interested in transpositions, we include the en-passant square 
only if an en-passant capture is possible, and we omit the half and
full move fields.
*/

void board_to_fen(char *string, int strict)
{
	int i, j, k;
	int n;
	board_t *p;
	
	strcpy(string, "");
	
	k = 0;
	for (j = 0; j < 8; j++)
	{
		n = 0;
		for (i = 0; i < 8; i++)
		{
			p = board[map[k]];
			if (p == &empty)
			{
				n++;
			}
			else
			{
				strcat(string, empty_fen[n]);
				n = 0;
				if (p->board_colour == WHITE_COLOUR)
					strcat(string, white_fen[p->board_kind]);
				else if (p->board_colour == BLACK_COLOUR)
					strcat(string, black_fen[p->board_kind]);
				else
					FATAL_ERROR("p->board_colour out of range")
			}
			k++;
		}
		strcat(string, empty_fen[n]);
		n = 0;
		if (j < 7)
			strcat(string, "/");
		else
			strcat(string, " ");
	}
	if (my_colour == WHITE_COLOUR)
		strcat(string, "w ");
	else if (my_colour == BLACK_COLOUR)
		strcat(string, "b ");
	else
		FATAL_ERROR("my_colour out of range")
		
	/*
	castling
	*/
	
	n = 0;
	if (!board[E1]->board_moved)
	{
		if (!board[H1]->board_moved)
		{
			strcat(string, "K");
			n++;
		}
		if (!board[A1]->board_moved)
		{
			strcat(string, "Q");
			n++;
		}
	}
	if (!board[E8]->board_moved)
	{
		if (!board[H8]->board_moved)
		{
			strcat(string, "k");
			n++;
		}
		if (!board[A1]->board_moved)
		{
			strcat(string, "q");
			n++;
		}
	}
	if (n == 0)
		strcat(string, "- ");
	else
		strcat(string, " ");
	
	/*
	en passant
	*/
	
	if (hack_ep == INVALID)
		strcat(string, "-");
	else
	{
		int ep;
		
		if (strict)
		{
			ep = TRUE;
		}
		else
		{
			ep = FALSE;
			j = hack_ep - 1;
			if ((board[j]->board_colour == my_colour) and (board[j]->board_kind == PAWN))
				ep = TRUE;
			j = hack_ep + 1;
			if ((board[j]->board_colour == my_colour) and (board[j]->board_kind == PAWN))
				ep = TRUE;
		}
		if (ep)
		{
			strcat(string, file_fen[column[hack_ep]]);
			if (row[hack_ep] == 3)
				strcat(string, "3");
			else if (row[hack_ep] == 4)
				strcat(string, "6");
			else
				FATAL_ERROR("hack_ep out of range")
		}
		else
			strcat(string, "-");
	}
	
	/*
	remaining fields are dummy for the moment
	*/
	
	if (strict)
		strcat(string, " 0 0");
}

/*
return_cur_key returns the current hash key.
*/

local hash_key_t return_cur_key(void);
local hash_key_t return_cur_key(void)
{
	int icolour, ikind;
	hash_key_t key;
	board_t *p;
	
	CLEAR_KEY(key)
	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		for (ikind = PAWN; ikind <= KING; ikind++)
		{
			for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
			{
				if (p->board_where != INVALID)
					UPDATE_KEY(key, p->board_key[p->board_where])
			}
		}
	}
	return(key);
}

/*
write_board writes the board to a file.
*/

void write_board(FILE *f)
{
	int i, j, k;
	
	k = 0;
	for (j = 0; j < 8; j++)
	{
		for (i = 0; i < 8; i++)
		{
			if (board[map[k]] == &empty)
				fprintf(f, "..");
			else
				fprintf(f, "%c%c",
					nota_colour[board[map[k]]->board_colour],
					nota_kind[board[map[k]]->board_kind]);
			k++;
		}
		fprintf(f, "\n");
	}
}

/*
display_list displays the lists of pieces. Captured pieces are
indicated by ??
*/

void display_list(void)
{
	int icolour;
	int ikind;
	board_t *p;

	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		my_printf("%s%d", colour_as_string[icolour], tot_count[icolour]);
		for (ikind = PAWN; ikind <= KING; ikind++)
		{
			my_printf(" %d%c", piece_count[icolour][ikind], nota_kind[ikind]);
			for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
			{
				if (p->board_where == INVALID)
					my_printf(" ??");
				else
					my_printf(" %s", GET_NOTA(p->board_where));
			}
		}
		my_printf("\n");
	}
}

/*
display_occup displays the occupation numbers for the columns.
*/

void display_occup(void)
{
	int icolour;
	int icolumn;
	
	for (icolour = BLACK_COLOUR; icolour >= WHITE_COLOUR; icolour--)
	{
		my_printf("%s", colour_as_string[icolour]);
		for (icolumn = 0; icolumn < 8; icolumn++)
			my_printf(" %d", bit_count[occup[icolour][icolumn]]);
		my_printf("\n");
	}
}

/*
display_attack displays the attack table.
*/

void display_attack(void)
{
	int i, j, k;
	
	k = 0;
	for (j = 0; j < 8; j++)
	{
		for (i = 0; i < 8; i++)
		{
			my_printf(" %2d", attack_board[map[k]]);
			k++;
		}
		my_printf("\n");
	}
}

/*
return_mat_score calculates the current material value
*/

local int return_mat_score(int);
local int return_mat_score(int my_colour)
{
	int sum;
	int ikind;
	board_t *p;
	
	sum = 0;
	for (ikind = PAWN; ikind <= QUEEN; ikind++)
	{
		for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++)
			if (p->board_where != INVALID) sum += p->board_value;
	}
	return(sum);
}

/*
update_key updates the hash key with ep rights and castling rights.
This makes the position unique.
*/

local void update_key(int, hash_key_t *);
local void update_key(int your_colour, hash_key_t *key)
{
	int j;
	
	if (hack_ep != INVALID)
	{
		/*
		Pawn to the left?
		*/
		
		j = hack_ep - 1;
		if ((board[j]->board_colour == your_colour) and (board[j]->board_kind == PAWN))
			UPDATE_KEY(*key, ep_key[j])
		
		/*
		Pawn to the right?
		*/
		
		j = hack_ep + 1;
		if ((board[j]->board_colour == your_colour) and (board[j]->board_kind == PAWN))
			UPDATE_KEY(*key, ep_key[j])
	}
	if (!board[E1]->board_moved)
	{
		if (!board[H1]->board_moved)
			UPDATE_KEY(*key, castle_key[H1])
		if (!board[A1]->board_moved)
			UPDATE_KEY(*key, castle_key[A1])
	}
	if (!board[E8]->board_moved)
	{
		if (!board[H8]->board_moved)
			UPDATE_KEY(*key, castle_key[H8])
		if (!board[A1]->board_moved)
			UPDATE_KEY(*key, castle_key[A8])
	}
}

/*
check_lists checks whether the lists of pieces are still correct and returns
an error code if there is something wrong. Returns 0 if the lists are OK.
*/

local int check_list(void);
local int check_list(void)
{
	int i;
	int icolour;
	board_t *p;
	int ncolour, nkind;
	int ikind;
	
	for (i = 0; i < 64; i++)
		if ((board[map[i]] != &empty) and (board[map[i]]->board_where != map[i])) return(1);
	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		/*
		ncolour is the total number of pieces of colour icolour.
		*/
		
		ncolour = 0;
		for (ikind = PAWN; ikind <= KING; ikind++)
		{
			/*
			nkind is the total number of pieces of kind ikind.
			*/
			
			nkind = 0;
			for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
			{
				if (p->board_where != INVALID)
				{
					if (board[p->board_where] != p) return(2);
					nkind++;
				}
			}
			if (nkind != piece_count[icolour][ikind]) return(3);
			ncolour += nkind;
		}
		if (ncolour != tot_count[icolour]) return(4);
	}
	return(0);
}

/*
compact_lists removes captured pieces from the lists. 
Is called before we start thinking about a position.
*/

void compact_lists(void)
{
	int icolour, ikind;
	board_t *p;
	int err;
	int n, i;
	board_t temp_list[LIST_MAX];
	
	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		/*
		We skip the king of course.
		*/
		
		for (ikind = PAWN; ikind <= QUEEN; ikind++)
		{
			/*
			Copy all pieces to temp_list.
			*/
			
			n = 0;
			for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
			{
				if (p->board_where != INVALID)
				{
					BUG(n >= LIST_MAX)
					temp_list[n] = *p;
					n++;
				}
			}
			
			/*
			Copy back.
			*/
			
			p = list[icolour][ikind];
			for (i = 0; i < n; i++)
			{
				*p = temp_list[i];
				board[p->board_where] = p;
				p++;
			}
			
			/*
			Remaining entries are invalid.
			*/
			
			for ( ; i < LIST_MAX; i++)
			{
				p->board_kind = p->board_where = INVALID;
				p++;
			}
		}
	}
	if ((err = check_list()) > 0)
	{
		my_printf("err=%d\n", err);
		display_board();
		display_list();
		FATAL_ERROR("list error")
	}
}

/*
fen_to_board sets the board
*/

void fen_to_board(char *string, int strict)
{
	int i, j, k, n;
	int icolour, ikind;
	board_t *p;
	int err;
	char *sync;

	hack_ep = INVALID;

	/*
	Initialize the boundary and the board.
	*/
	
	for (i = 0; i < BOARD_MAX; i++)
		board[i] = &side;
	for (i = 0; i < 64; i++)
		board[map[i]] = &empty;
	
	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		tot_count[icolour] = 0;
		for (ikind = PAWN; ikind <= KING; ikind++)
		{
			/*
			Initialize all entries.
			*/
			
			p = list[icolour][ikind];
			for (i = 0; i < LIST_MAX; i++)
			{
				p->board_colour = icolour;
				p->board_kind = INVALID;
				p->board_value = value[ikind];
				p->board_where = INVALID;
				p->board_moved = TRUE;
				p->board_captured = FALSE;
				p->board_key = hash_key[icolour][ikind];
				p->board_shift = shift[ikind];
				p->board_dir = dir[icolour][ikind];
				p->board_count = piece_count[icolour] + ikind;
				p++;
			}
			piece_count[icolour][ikind] = 0;
		}
	}
	
	k = 0;
	sync = start_board;
	for (i = 0; i < 8; i++)
	{
		j = 0;
		while (j < 8)
		{
			for (n = 8; n > 0; n--)
			{
				if (*string == *empty_fen[n]) break;
			}
			if (n > 0)
			{
				j += n;
				k += n;
				sync += n;
				string++;
				continue;
			}
			icolour = INVALID;
			for (ikind = PAWN; ikind <= KING; ikind++)
			{
				if (*string == *black_fen[ikind])
				{
					icolour = BLACK_COLOUR;
					break;
				}
				if (*string == *white_fen[ikind])
				{
					icolour = WHITE_COLOUR;
					break;
				}
			}
			BUG(icolour == INVALID)
			p = list[icolour][ikind] + piece_count[icolour][ikind];
			p->board_kind = ikind;
			p->board_where = map[k];
			board[map[k]] = p;
			piece_count[icolour][ikind]++;
			tot_count[icolour]++;

			/*
			We assume a piece has not moved if it is still on it's
			starting position.
			*/
			
			if (*string == *sync)
				p->board_moved = FALSE;
			else
				p->board_moved = TRUE;
			j++;
			k++;
			string++;
			sync++;
		}
		if (i < 7)
		{
			BUG(*string != '/');
			BUG(*sync != '/');
		}
		string++;
		sync++;
	}

	if (*string == 'w')
		WHITE_TO_MOVE
	else if (*string == 'b')
		BLACK_TO_MOVE
	else
		FATAL_ERROR("colour out of range!")
	string++;
	
	BUG(*string != ' ');
	string++;
	
	/*
	possible castling rights
	*/
	
	if (*string == '-')
	{
		board[E1]->board_moved = TRUE;
		board[A1]->board_moved = TRUE;
		board[H1]->board_moved = TRUE;
		board[E8]->board_moved = TRUE;
		board[A8]->board_moved = TRUE;
		board[H8]->board_moved = TRUE;
		string++;
	}
	else if (*string == '?')
	{
		if (strict) FATAL_ERROR("invalid castle string");
		string++;
	}
	else
	{
		while(*string != ' ')
		{
			if (*string == 'K')
			{
				board[E1]->board_moved = FALSE;
				board[H1]->board_moved = FALSE;
			}
			else if (*string == 'Q')
			{
				board[E1]->board_moved = FALSE;
				board[A1]->board_moved = FALSE;
			}
			else if (*string == 'k')
			{
				board[E8]->board_moved = FALSE;
				board[H8]->board_moved = FALSE;
			}
			else if (*string == 'q')
			{
				board[E8]->board_moved = FALSE;
				board[A8]->board_moved = FALSE;
			}
			else
				FATAL_ERROR("fen error")
			string++;
		}
	}
	
	BUG(*string != ' ')
	string++;
	
	if (*string == '-')
	{
		hack_ep = INVALID;
		string++;
	}
	else
	{
		for (k = 0; k < 64; k++)
		{
			if (strncmp(string, nota_table[k], 2) == 0)
			{
				hack_ep = map[k] + pawn_dir[your_colour][0];
				break;
			}
		}
		BUG(k == 64)
		string++;
		string++;
	}
	
	if (strict)
	{
		BUG(*string != ' ')
		string++;
	}
	
	BUG(piece_count[WHITE_COLOUR][KING] != 1)
	BUG(piece_count[BLACK_COLOUR][KING] != 1)
	
	capt_top = capt_stack;
	
	/*
	Set node.
	*/

	node = node_stack;
	node->node_key = return_cur_key();
	node->node_total_key = node->node_key;
	update_key(your_colour, &(node->node_total_key));
	node->node_mat[WHITE_COLOUR] = return_mat_score(WHITE_COLOUR);
	node->node_mat[BLACK_COLOUR] = return_mat_score(BLACK_COLOUR);
	node->node_reverse = FALSE;
	node->node_hack_ep = INVALID;
	
	if ((err = check_list()) > 0)
	{
		my_printf("err=%d\n", err);
		display_board();
		display_list();
		FATAL_ERROR("list error")
	}
}

/*
attacked returns TRUE if square i is attacked by your_colour.
*/

local int attacked(int, int);
local int attacked(int i, int your_colour)
{
	int *d;
	board_t *p;
	int j;
	
	p = board[i - pawn_dir[your_colour][1]];
	if ((p->board_colour == your_colour) and (p->board_kind == PAWN))
		return(TRUE);
	p = board[i - pawn_dir[your_colour][2]];
	if ((p->board_colour == your_colour) and (p->board_kind == PAWN))
		return(TRUE);
	for (d = knight_dir[your_colour]; *d != 0; d++)
	{
		p = board[i - *d];
		if ((p->board_colour == your_colour) and (p->board_kind == KNIGHT))
			return(TRUE);
	}
	for (d = bishop_dir[your_colour]; *d != 0; d++)
	{
		j = i - *d;
		p = board[j];
		if ((p->board_colour == your_colour) and (p->board_kind == KING))
			return(TRUE);
		while(board[j] == &empty) j -= *d;
		p = board[j];
		if ((p->board_colour == your_colour) and
		    ((p->board_kind == BISHOP) or (p->board_kind == QUEEN)))
			return(TRUE);
	}
	for (d = rook_dir[your_colour]; *d != 0; d++)
	{
		j = i - *d;
		p = board[j];
		if ((p->board_colour == your_colour) and (p->board_kind == KING))
			return(TRUE);
		while(board[j] == &empty) j -= *d;
		p = board[j];
		if ((p->board_colour == your_colour) and
		    ((p->board_kind == ROOK) or (p->board_kind == QUEEN)))
			return(TRUE);
	}
	return(FALSE);
}

int i_am_check(int my_colour, int your_colour)
{
	return(attacked(king_list[my_colour]->board_where, your_colour));
}

/*
gen_moves generates the moves for the piece on square p.
*/

local void gen_moves(int, int, board_t *);
local void gen_moves(int my_colour, int your_colour, board_t *p)
{
	int j;
	int *d;
	int irow;
	
	if (p->board_kind == PAWN)
	{
		/*
		Pawn captures.
		*/
		
		j = p->board_where + p->board_dir[1];
		irow = row[j];
		if (board[j]->board_colour == your_colour)
		{
			if (irow == prom_row[my_colour])
			{
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_QUEEN ), p->board_where, j)
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_ROOK  ), p->board_where, j)
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_BISHOP), p->board_where, j)
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_KNIGHT), p->board_where, j)
			}
			else
			{
				ADD_MOVE(capt_top, CAPT, p->board_where, j)
			}
		}
		j = p->board_where + p->board_dir[2];
		irow = row[j];
		if (board[j]->board_colour == your_colour)
		{
			if (irow == prom_row[my_colour])
			{
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_QUEEN ), p->board_where, j)
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_ROOK  ), p->board_where, j)
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_BISHOP), p->board_where, j)
				ADD_MOVE(capt_top, (PROM_CAPT | PROM_KNIGHT), p->board_where, j)
			}
			else
			{
				ADD_MOVE(capt_top, CAPT, p->board_where, j)
			}
		}
		
		/*
		non-capturing moves for pawn.
		*/
		
		j = p->board_where + p->board_dir[0];
		if (board[j] == &empty)
		{
			irow = row[j];
			if (irow == prom_row[my_colour])
			{
				ADD_MOVE(move_top, (PROM_NORM | PROM_QUEEN ), p->board_where, j)
				ADD_MOVE(move_top, (PROM_NORM | PROM_ROOK  ), p->board_where, j)
				ADD_MOVE(move_top, (PROM_NORM | PROM_BISHOP), p->board_where, j)
				ADD_MOVE(move_top, (PROM_NORM | PROM_KNIGHT), p->board_where, j)
			}
			else
			{
				ADD_MOVE(move_top, NORM, p->board_where, j)
			}
			
			/*
			double pawn move
			*/
			
			if (!p->board_moved)
			{
				j += p->board_dir[0];
				if (board[j] == &empty)
				{
					ADD_MOVE(move_top, DOUBLE, p->board_where, j)
				}
			}
		}
	}
	else
	{
		/*
		Other pieces.
		*/
		
		for (d = p->board_dir; *d != 0; d++)
		{
			j = p->board_where;
			while(TRUE)
			{
				/*
				Try next square.
				*/
				
				j += *d;
				if (board[j] == &empty)
				{
					ADD_MOVE(move_top, NORM, p->board_where, j)
					if (!p->board_shift) break;
				}
				else
				{
					/*
					May be capture.
					*/
					
					if (board[j]->board_colour == your_colour)
					{
						ADD_MOVE(capt_top, CAPT, p->board_where, j)
					}
					break;
				}
			}
		}
	}
}

/*
gen_special_moves generates en passant and short or long castle.
*/

local void gen_special_moves(int, int);
local void gen_special_moves(int my_colour, int your_colour)
{
	int j;
	
	if (hack_ep != INVALID)
	{
		/*
		Pawn to the left?
		*/
		
		j = hack_ep - 1;
		if ((board[j]->board_colour == my_colour) and (board[j]->board_kind == PAWN))
		{
			ADD_MOVE(capt_top, EP, j, hack_ep + pawn_dir[my_colour][0])
		}
		
		/* 
		Pawn to the right?
		*/
		
		j = hack_ep + 1;
		if ((board[j]->board_colour == my_colour) and (board[j]->board_kind == PAWN))
		{
			ADD_MOVE(capt_top, EP, j, hack_ep + pawn_dir[my_colour][0])
		}
	}
	if (my_colour == WHITE_COLOUR)
	{
		/*
		empty square has board_moved set to TRUE. Dirty trick eh?
		*/
		
		if (!board[E1]->board_moved)
		{
			if (!board[H1]->board_moved)
			{
				if ((board[F1] == &empty) and (board[G1] == &empty) and
				    !attacked(E1, your_colour) and !attacked(F1, your_colour) and !attacked(G1, your_colour))
				{
					ADD_MOVE(capt_top, SHORT, 0, 0)
				}
			}
			if (!board[A1]->board_moved)
			{
				if ((board[D1] == &empty) and (board[C1] == &empty) and
				    (board[B1] == &empty) and
				    !attacked(E1, your_colour) and !attacked(D1, your_colour) and !attacked(C1, your_colour))
				{
					ADD_MOVE(capt_top, LONG, 0, 0)
				}
			}
		}
	}
	else
	{
		if (!board[E8]->board_moved)
		{
			if (!board[H8]->board_moved)
			{
				if ((board[F8] == &empty) and (board[G8] == &empty) and
				    !attacked(E8, your_colour) and !attacked(F8, your_colour) and !attacked(G8, your_colour))
				{  
					ADD_MOVE(capt_top, SHORT, 0, 0)
				}
			}
			if (!board[A8]->board_moved)
			{
				if ((board[D8] == &empty) and (board[C8] == &empty) and
				    (board[B8] == &empty) and
				    !attacked(E8, your_colour) and !attacked(D8, your_colour) and !attacked(C8, your_colour))
				{
					ADD_MOVE(capt_top, LONG, 0, 0)
				}
			}
		}
	}
}

/*
gen_all_moves generates all moves for a colour.
*/

int gen_all_moves(void)
{
	int ikind, imove;
	board_t *p;
	int nmoves;
	move_t *base;
	
	move_top = move_stack;
	base = capt_top;
	for (ikind = PAWN; ikind <= KING; ikind++)
	{
		for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++)
		{
			if (p->board_where == INVALID) continue;
			gen_moves(my_colour, your_colour, p);
		}
	}
	gen_special_moves(my_colour, your_colour);
	nmoves = move_top - move_stack;
	for (imove = 0; imove < nmoves; imove++)
		*capt_top++ = move_stack[imove];
	nmoves = capt_top - base;
#if DEBUG_LEVEL > 0
	BUG((capt_top - capt_stack) >= CAPT_MAX)
#endif
	return(nmoves);
}

/*
gen_capts generates all captures to a square. Same as attacked routine,
except ADD_MOVE is added.
*/

local int gen_capts(int, int);
local int gen_capts(int i, int your_colour)
{
	int *d;
	board_t *p;
	int j;
	move_t *base;
	
	base = capt_top;
	j = i - pawn_dir[your_colour][1];
	p = board[j];
	if ((p->board_colour == your_colour) and (p->board_kind == PAWN))
	{
		if (row[i] == prom_row[your_colour])
		{
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_QUEEN ), j, i)
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_ROOK  ), j, i)
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_BISHOP), j, i)
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_KNIGHT), j, i)
		}
		else
			ADD_MOVE(capt_top, CAPT, j, i)
	}
	j = i - pawn_dir[your_colour][2];
	p = board[j];
	if ((p->board_colour == your_colour) and (p->board_kind == PAWN))
	{
		if (row[i] == prom_row[your_colour])
		{
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_QUEEN ), j, i)
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_ROOK  ), j, i)
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_BISHOP), j, i)
			ADD_MOVE(capt_top, (PROM_CAPT | PROM_KNIGHT), j, i)
		}
		else
			ADD_MOVE(capt_top, CAPT, j, i)
	}
	for (d = knight_dir[your_colour]; *d != 0; d++)
	{
		j = i - *d;
		p = board[j];
		if ((p->board_colour == your_colour) and (p->board_kind == KNIGHT)) ADD_MOVE(capt_top, CAPT, j, i)
	}
	for (d = bishop_dir[your_colour]; *d != 0; d++)
	{
		j = i - *d;
		p = board[j];
		if ((p->board_colour == your_colour) and (p->board_kind == KING)) ADD_MOVE(capt_top, CAPT, j, i)
		while(board[j] == &empty) j -= *d;
		p = board[j];
		if ((p->board_colour == your_colour) and
		    ((p->board_kind == BISHOP) or (p->board_kind == QUEEN))) ADD_MOVE(capt_top, CAPT, j, i)
	}
	for (d = rook_dir[your_colour]; *d != 0; d++)
	{
		j = i - *d;
		p = board[j];
		if ((p->board_colour == your_colour) and (p->board_kind == KING)) ADD_MOVE(capt_top, CAPT, j, i)
		while(board[j] == &empty) j -= *d;
		p = board[j];
		if ((p->board_colour == your_colour) and
		    ((p->board_kind == ROOK) or (p->board_kind == QUEEN))) ADD_MOVE(capt_top, CAPT, j, i)
	}
	return(capt_top - base);
}

/*
Chess do_move routines always seem to get a bit messy when you add features
like piece lists and piece counts..
*/

void do_move(move_t m)
{
/*
DO_CAPT removes a piece on SQUARE B. 
*/

#define DO_CAPT(B)\
	{\
		p = board[B];\
		if (p->board_colour != your_colour) {display_board(); display_list(); my_printf("ep=%s %s\n", GET_NOTA(hack_ep), colour_as_string[my_colour]);}\
		BUG(p->board_kind == KING)\
		node->capt = p;\
		p->board_where = INVALID;\
		UPDATE_KEY(node->node_key, p->board_key[B])\
		node->node_mat[your_colour] -= p->board_value;\
		tot_count[your_colour]--;\
		(*p->board_count)--;\
	}

/*
DO_MOVE moves a piece from square A to square B.
*/

#define DO_MOVE(A, B)\
	{\
		p = board[A];\
		UPDATE_KEY(node->node_key, p->board_key[A])\
		UPDATE_KEY(node->node_key, p->board_key[B])\
		p->board_where = B;\
		node->node_moved = p->board_moved;\
		node->node_captured = p->board_captured;\
		p->board_moved = TRUE;\
		board[A] = &empty;\
		board[B] = p;\
	}

	board_t *p;
	long kind;
	int from, to;
	int ep;
#if DEBUG_LEVEL > 0
	int err;
	hash_key_t temp_key;
#endif
	
	/*
	Save move and hack_ep for undo move
	*/
	
	node->node_move = m;
	node->node_hack_ep = hack_ep;
	hack_ep = INVALID;
	
	/*
	Copy incrementally updated information to next node.
	*/
	
	node[1].node_key = node->node_key;
	node[1].node_mat[WHITE_COLOUR] = node->node_mat[WHITE_COLOUR];
	node[1].node_mat[BLACK_COLOUR] = node->node_mat[BLACK_COLOUR];
	node++;
	inode++;
	
	node->node_reverse = FALSE;
	UNPACK_MOVE(m, kind, from, to)
	if (kind == NULL_MOVE) goto label_null;
	
	/*
	Another cheap check.
	*/
	
	if ((kind != SHORT) and (kind != LONG))
	{
		BUG(column[from] == INVALID)
		BUG(column[to] == INVALID)
	}
	if (kind == EP)
	{
		ep = to - pawn_dir[my_colour][0];
		DO_CAPT(ep)
		board[ep] = &empty;
		DO_MOVE(from, to)
	}
	else if (kind & PROM_CAPT)
	{
		DO_CAPT(to)
		
		/*
		Remove pawn
		*/
		
		p = board[from];
		node->prom = p;
		board[from] = &empty;
		p->board_where = INVALID;
		(*p->board_count)--;
		UPDATE_KEY(node->node_key, p->board_key[from])
		node->node_mat[my_colour] -= p->board_value;
		
		/*
		Add piece. Search end of list first.
		*/
		
		if (kind & PROM_QUEEN)
		{
			for (p = queen_list[my_colour]; p->board_kind == QUEEN; p++);
			p->board_kind = QUEEN;
		}
		else if (kind & PROM_ROOK)
		{
			for (p = rook_list[my_colour]; p->board_kind == ROOK; p++);
			p->board_kind = ROOK;
		}
		else if (kind & PROM_BISHOP)
		{
			for (p = bishop_list[my_colour]; p->board_kind == BISHOP; p++);
			p->board_kind = BISHOP;
		}
		else
		{
			for (p = knight_list[my_colour]; p->board_kind == KNIGHT; p++);
			p->board_kind = KNIGHT;
		}
		p->board_where = to;
		(*p->board_count)++;
		board[to] = p;
		UPDATE_KEY(node->node_key, p->board_key[to])
		node->node_mat[my_colour] += p->board_value;
	}
	else if (kind == SHORT)
	{
		if (my_colour == WHITE_COLOUR)
		{
			DO_MOVE(E1, G1)
			DO_MOVE(H1, F1)
		}
		else
		{
			DO_MOVE(E8, G8)
			DO_MOVE(H8, F8)
		}
	}
	else if (kind == LONG)
	{
		if (my_colour == WHITE_COLOUR)
		{
			DO_MOVE(E1, C1)
			DO_MOVE(A1, D1)
		}
		else
		{
			DO_MOVE(E8, C8)
			DO_MOVE(A8, D8)
		}
	}
	else if (kind & PROM_NORM)
	{
		/*
		Remove pawn
		*/

		p = board[from];
		node->prom = p;
		board[from] = &empty;
		p->board_where = INVALID;
		(*p->board_count)--;
		UPDATE_KEY(node->node_key, p->board_key[from])
		node->node_mat[my_colour] -= p->board_value;
		
		/*
		Add piece
		*/
		
		if (kind & PROM_QUEEN)
		{
			for (p = queen_list[my_colour]; p->board_kind == QUEEN; p++);
			p->board_kind = QUEEN;
		}
		else if (kind & PROM_ROOK)
		{
			for (p = rook_list[my_colour]; p->board_kind == ROOK; p++);
			p->board_kind = ROOK;
		}
		else if (kind & PROM_BISHOP)
		{
			for (p = bishop_list[my_colour]; p->board_kind == BISHOP; p++);
			p->board_kind = BISHOP;
		}
		else
		{
			for (p = knight_list[my_colour]; p->board_kind == KNIGHT; p++);
			p->board_kind = KNIGHT;
		}
		p->board_where = to;
		(*p->board_count)++;
		board[to] = p;
		UPDATE_KEY(node->node_key, p->board_key[to])
		node->node_mat[my_colour] += p->board_value;
	}
	else
	{
		if (kind == CAPT)
			DO_CAPT(to)
		DO_MOVE(from, to)
		if (kind == CAPT)
			p->board_captured = TRUE;
		else if (kind == DOUBLE)
			hack_ep = to;
		else
			node->node_reverse = TRUE;
	}
	
	label_null:
	node->node_total_key = node->node_key;
	update_key(your_colour, &(node->node_total_key));
#if DEBUG_LEVEL > 0
	if ((err = check_list()) > 0)
	{
		my_printf("%s err=%d\n", colour_as_string[my_colour], err);
		move_to_string(m);
		my_printf("%s\n", move_string);
		display_board();
		display_list();
		FATAL_ERROR("do list error")
	}
	temp_key = return_cur_key();
	if (!EQUAL_KEY(temp_key, node->node_key))
	{
		my_printf("%s\n", colour_as_string[my_colour]);
		move_to_string(m);
		my_printf("%s\n", move_string);
		display_board();
		display_list();
		FATAL_ERROR("do key error")
	}
	if (node->node_mat[WHITE_COLOUR] != return_mat_score(WHITE_COLOUR))
	{
		my_printf("%s\n", colour_as_string[my_colour]);
		move_to_string(m);
		my_printf("%s\n", move_string);
		display_board();
		display_list();
		FATAL_ERROR("do white mat error")
	}
	if (node->node_mat[BLACK_COLOUR] != return_mat_score(BLACK_COLOUR))
	{
		my_printf("%s\n", colour_as_string[my_colour]);
		move_to_string(m);
		my_printf("%s\n", move_string);
		display_board();
		display_list();
		FATAL_ERROR("do black mat error")
	}
#endif
	if (my_colour == WHITE_COLOUR)
		BLACK_TO_MOVE
	else
		WHITE_TO_MOVE
#undef DO_CAPT
#undef DO_MOVE
}

void undo_move(void)
{
#define MOVE(A, B)\
	{\
		p = board[A];\
		p->board_where = B;\
		p->board_moved = node->node_moved;\
		p->board_captured = node->node_captured;\
		board[A] = &empty;\
		board[B] = p;\
	}

	board_t *p;
	long kind;
	int from, to;
	int ep;
#if DEBUG_LEVEL > 0
	int err;
#endif

	if (my_colour == WHITE_COLOUR)
		BLACK_TO_MOVE
	else
		WHITE_TO_MOVE

	/*
	We have to peek in the previous node for the move.
	*/
	
	UNPACK_MOVE(node[-1].node_move, kind, from, to)
	if (kind == NULL_MOVE) goto label_null;
	if ((kind != SHORT) and (kind != LONG))
	{
		BUG(column[from] == INVALID)
		BUG(column[to] == INVALID)
	}
	if (kind == EP)
	{
		ep = to - pawn_dir[my_colour][0];
		p = board[ep] = node->capt;
		p->board_where = ep;
		(*p->board_count)++;
		MOVE(to, from)
		tot_count[your_colour]++;
	}
	else if (kind & PROM_CAPT)
	{
		p = board[to];
		p->board_kind = p->board_where = INVALID;
		(*p->board_count)--;
		
		p = board[to] = node->capt;
		p->board_where = to;
		tot_count[your_colour]++;
		(*p->board_count)++;

		p = board[from] = node->prom;
		p->board_where = from;
		(*p->board_count)++;
	}
	else if (kind == SHORT)
	{
		if (my_colour == WHITE_COLOUR)
		{
			MOVE(F1, H1)
			MOVE(G1, E1)
		}
		else
		{
			MOVE(F8, H8)
			MOVE(G8, E8)
		}
	}
	else if (kind == LONG)
	{
		if (my_colour == WHITE_COLOUR)
		{
			MOVE(D1, A1)
			MOVE(C1, E1)
		}
		else
		{
			MOVE(D8, A8)
			MOVE(C8, E8)
		}
	}
	else if (kind & PROM_NORM)
	{
		p = board[to];
		p->board_kind = p->board_where = INVALID;
		(*p->board_count)--;
		board[to] = &empty;
		p = board[from] = node->prom;
		p->board_where = from;
		(*p->board_count)++;
	}
	else
	{
		MOVE(to, from)
		if (kind == CAPT)
		{
			p = board[to] = node->capt;
			p->board_where = to;
			tot_count[your_colour]++;
			(*p->board_count)++;
		}
	}
	label_null:
	node--;
	inode--;
	BUG(inode < 0)
	hack_ep = node->node_hack_ep;
#if DEBUG_LEVEL > 0
	if ((err = check_list()) > 0)
	{
		my_printf("%s err=%d\n", colour_as_string[my_colour], err);
		move_to_string(node->node_move);
		my_printf("%s\n", move_string);
		display_board();
		display_list();
		FATAL_ERROR("undo list error")
	}
#endif
}

/*
move_to_san_string converts a move to Short Algebraic Notation
*/

void move_to_san_string(move_t m, int strict)
{
	int check;
	long kind;
	int from, to;
	char c;
	int ikind;
	board_t *p;
	int imove, nmoves;
	int other_kind, other_from, other_to;
	int nfound_col, nfound_row;
	int nlegal;
	move_t *base;
	
	/*
	We check if this is a legal move
	*/

	do_move(m);
	check = i_am_check(your_colour, my_colour);
	undo_move();
	if (check)
	{
		sprintf(move_string, "??");
		return;
	}
	
	UNPACK_MOVE(m, kind, from, to)
	
	/*
	First some unambigous cases.
	*/
	
	if (kind & EP)
	{
		sprintf(move_string, "%cx%s", GET_COL(from), GET_NOTA(to));
	}
	else if (kind & SHORT)
	{
		sprintf(move_string, "O-O");
	}
	else if (kind & LONG)
	{
		sprintf(move_string, "O-O-O");
	}
	else if (kind & PROM_CAPT)
	{
		if (kind & PROM_QUEEN)
			c = 'Q';
		else if (kind & PROM_ROOK)
			c = 'R';
		else if (kind & PROM_BISHOP)
			c = 'B';
		else
			c = 'N';
		sprintf(move_string, "%cx%s=%c", GET_COL(from), GET_NOTA(to), c);
	}
	else if (kind & PROM_NORM)
	{
		if (kind & PROM_QUEEN)
			c = 'Q';
		else if (kind & PROM_ROOK)
			c = 'R';
		else if (kind & PROM_BISHOP)
			c = 'B';
		else
			c = 'N';
		sprintf(move_string, "%s=%c", GET_NOTA(to), c);
	}
	else if (board[from]->board_kind == PAWN)
	{
		if (kind & CAPT)
		{
			sprintf(move_string, "%cx%s", GET_COL(from), GET_NOTA(to));
		}
		else
		{
			sprintf(move_string, "%s", GET_NOTA(to));
		}
	}
	else
	{
		/*
		Now the ambiguous cases. Ambigous moves are moves by the same kind of piece
		with the same 'to' field, so we have to generate some moves here.
		*/
		
		move_top = move_stack;
		base = capt_top;
		ikind = board[from]->board_kind;
		for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++)
		{
			if (p->board_where == INVALID) continue;
			gen_moves(my_colour, your_colour, p);
		}
		nmoves = move_top - move_stack;
		for (imove = 0; imove < nmoves; imove++)
			*capt_top++ = move_stack[imove];
		nmoves = capt_top - base;

		nfound_col = nfound_row = 0;
		
		for (imove = 0; imove < nmoves; imove++)
		{
			/*
			Skip the move itself.
			*/
			
			if (base[imove] == m) continue;
			
			/*
			Only legal moves
			*/
			
			do_move(base[imove]);
			check = i_am_check(your_colour, my_colour);
			undo_move();
			if (check) continue;
			
			UNPACK_MOVE(base[imove] , other_kind, other_from, other_to)
			if (other_to == to)
			{
				/* 
				This move seems to be ambiguous.
				Distinguish by file and by rank.
				*/
				
				if (column[other_from] != column[from])
					nfound_col++;
				else if (row[other_from] != row[from])
					nfound_row++;
				else
					FATAL_ERROR("from")
			}
		}
		capt_top = base;
		
		c = nota_kind[board[from]->board_kind];
		if ((nfound_col == 0) and (nfound_row == 0))
		{
			/*
			Not ambiguous.
			*/
			
			if (kind & CAPT)
				sprintf(move_string, "%cx%s", c, GET_NOTA(to));
			else
				sprintf(move_string, "%c%s", c, GET_NOTA(to));
		}	
		else if (nfound_col == 1)
		{
			if (kind & CAPT)
				sprintf(move_string, "%c%cx%s", c, GET_COL(from), GET_NOTA(to));
			else
				sprintf(move_string, "%c%c%s", c, GET_COL(from), GET_NOTA(to));
		}
		else if (nfound_row == 1)
		{
			if (kind & CAPT)
				sprintf(move_string, "%c%cx%s", c, GET_ROW(from), GET_NOTA(to));
			else
				sprintf(move_string, "%c%c%s", c, GET_ROW(from), GET_NOTA(to));
		}
		else
		{
			if (kind & CAPT)
				sprintf(move_string, "%c%sx%s", c, GET_NOTA(from), GET_NOTA(to));
			else
				sprintf(move_string, "%c%s%s", c, GET_NOTA(from), GET_NOTA(to));
		}
	}

	/*
	Finally add check or mate
	*/
	
	if (strict)
	{
		do_move(m);
		check = i_am_check(my_colour, your_colour);
		if (check)
		{
			/*
			Is this a mating move?
			*/
	
			base = capt_top;
			nlegal = nmoves = gen_all_moves();
			for (imove = 0; imove < nmoves; imove++)
			{
				do_move(base[imove]);
				check = i_am_check(your_colour, my_colour);
				undo_move();
				if (check) nlegal--;
			}
			capt_top = base;
			if (nlegal == 0)
				strcat(move_string, "#");
			else
				strcat(move_string, "+");
		}
		undo_move();
	}
}

int search_san(char *move)
{
	int imove;
	int nmoves;
	move_t *base;
	
	base = capt_top;
	nmoves = gen_all_moves();
	for (imove = 0; imove < nmoves; imove++)
	{
		move_to_san_string(base[imove], TRUE);
		if (strcmp(move, move_string) == 0) break;
	}
	if (imove == nmoves)
		return(INVALID);
	return(imove);
}

/*
Score simple endgames.
*/

local int return_endgame_score(int, int);
local int return_endgame_score(int my_colour, int your_colour)
{
	int drow, dcol, dking;
	int ikind;
	board_t *p;
	
	BUG((tot_count[WHITE_COLOUR] + tot_count[BLACK_COLOUR]) > 3)
	drow = row[king_list[my_colour]->board_where] - row[king_list[your_colour]->board_where];
	dcol = column[king_list[my_colour]->board_where] - column[king_list[your_colour]->board_where];
	dking = ABS(drow) + ABS(dcol);
	
	if (tot_count[your_colour] == 1)
	{
		if (tot_count[my_colour] == 1)
			return(0);
		
		/*
		Find piece.
		*/
		
		for (ikind = PAWN; ikind <= QUEEN; ikind++)
		{
			for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++)
				if (p->board_where != INVALID) goto foundx;
		}
		foundx:
		if (p->board_kind == PAWN)
			return(free_pawn_grad[my_colour][p->board_where] - 0);
		else if (p->board_kind == KNIGHT)
			return(0);
		else if (p->board_kind == BISHOP)
			return(0);
		else if (p->board_kind == ROOK)
			return(SCORE_ENDGAME - dking - king_alone_grad[king_list[your_colour]->board_where]);
		else if (p->board_kind == QUEEN)
			return(SCORE_ENDGAME - dking - king_alone_grad[king_list[your_colour]->board_where]);
		else
			FATAL_ERROR("board error")
	}
	else
	{
		/*
		Find piece.
		*/
		
		for (ikind = PAWN; ikind <= QUEEN; ikind++)
		{
			for (p = list[your_colour][ikind]; p->board_kind != INVALID; p++)
				if (p->board_where != INVALID) goto foundy;
		}
		foundy:
		if (p->board_kind == PAWN)
			return(0 - free_pawn_grad[your_colour][p->board_where]);
		else if (p->board_kind == KNIGHT)
			return(0);
		else if (p->board_kind == BISHOP)
			return(0);
		else if (p->board_kind == ROOK)
			return(dking + king_alone_grad[king_list[my_colour]->board_where] - SCORE_ENDGAME);
		else if (p->board_kind == QUEEN)
			return(dking + king_alone_grad[king_list[my_colour]->board_where] - SCORE_ENDGAME);
		else
			FATAL_ERROR("board error")
	}
	return(SCORE_INVAL);
}

local void set_attack_board(int, int);
local void set_attack_board(int my_colour, int your_colour)
{
	board_t *p;
	int ikind;
	int *d, j;
	int icolumn;
	
	memcpy(attack_board + A8, attack_zero + A8, (H1 - A8 + 1) * sizeof(int));
	for (icolumn = 0; icolumn < 8; icolumn++)
		occup[WHITE_COLOUR][icolumn] = occup[BLACK_COLOUR][icolumn] = 0;
	for (ikind = PAWN; ikind <= KING; ikind++)
	{
		for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++)
		{
			if (p->board_where == INVALID) continue;
			if (p->board_kind == PAWN)
			{
				occup[my_colour][column[p->board_where]] |= BIT(row[p->board_where]);
				j = p->board_where + pawn_dir[my_colour][1];
				attack_board[j]++;
				j = p->board_where + pawn_dir[my_colour][2];
				attack_board[j]++;
			}
			else if (!p->board_shift)
			{
				for (d = p->board_dir; *d != 0; d++)
				{
					j = p->board_where + *d;
					attack_board[j]++;
				}
			}
			else
			{
				for (d = p->board_dir; *d != 0; d++)
				{
					for (j = p->board_where + *d; board[j] == &empty; j += *d)
						attack_board[j]++;
					attack_board[j]++;
				}
			}
		}
		
		/*
		Same for your colour
		*/
		
		for (p = list[your_colour][ikind]; p->board_kind != INVALID; p++)
		{
			if (p->board_where == INVALID) continue;
			if (p->board_kind == PAWN)
			{
				occup[your_colour][column[p->board_where]] |= BIT(row[p->board_where]);
				j = p->board_where + pawn_dir[your_colour][1];
				attack_board[j]--;
				j = p->board_where + pawn_dir[your_colour][2];
				attack_board[j]--;
			}
			else if (!p->board_shift)
			{
				for (d = p->board_dir; *d != 0; d++)
				{
					j = p->board_where + *d;
					attack_board[j]--;
				}
			}
			else
			{
				for (d = p->board_dir; *d != 0; d++)
				{
					for (j = p->board_where + *d; board[j] == &empty; j += *d)
						attack_board[j]--;
					attack_board[j]--;
				}
			}
		}
	}
}

/*
set_score sets the various positional parameters. Note: the assignment truncates.
This causes no problem when optimizing evaluation function parameters.
If the best value of for example weak_queen_score would be 10, than
y[0] would converge to a value between 10.0 and 10.99 if truncation is used,
but to a value between 9.5 and 10.49 if normal rounding was used.
*/

void set_score(double *y)
{
	weak_queen_score = y[0];
	weak_rook_score = y[1];
	weak_bishop_score = y[2];
	weak_knight_score = y[3];
	isolated_pawn_score = y[4];
	king_danger_score = y[5];
	double_pawn_score = y[6];
	lagged_pawn_score = y[7];
	bishop_blocks_pawn_score = y[8];
	bishop_not_developed_score = y[9];
	knight_not_developed_score = y[10];
	weak_pawn_attack_score = y[11];
	pin_score = y[12];
	rook_on_half_open_file_score = y[13];
	knight_strong_score = y[14];
	bishop_strong_score = y[15];
	rook_on_open_file_score = y[16];
	rook_on_7th_rank_score = y[17];
	long_castle_score = y[18];
	short_castle_score = y[19];
	knight_can_move_score = y[20];
	bishop_can_move_score = y[21];
	rook_can_move_score = y[22];
}

/*
return_pawn_pos_score analyzes the pawn structure. pawn_board is set
if the pawn is weak.
*/

local int return_pawn_pos_score(int, int, int);
local int return_pawn_pos_score(int my_colour, int your_colour, int verbose)
{
	board_t *p;
	int icolumn, irow;
	int sum = 0;
	int n;
	int tst;
	
	for (p = pawn_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		sum += pawn_visit[my_colour][p->board_where];
		pawn_board[p->board_where] = FALSE;
		icolumn = column[p->board_where];
		irow = row[p->board_where];
		if (bit_count[occup[my_colour][icolumn]] > 1)
		{
			if (verbose)
				my_printf("%spawn on %s is doubled\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
			sum += double_pawn_score;
		}
		
		/*
		For white: we have a free pawn if the row is greater than or equal to
		the highest bit of the occupation numbers of the current, left and right row. 
		*/
		
		if (icolumn == 0)
			n = occup[your_colour][1];
		else if (icolumn == 7)
			n = occup[your_colour][6];
		else
			n = (occup[your_colour][icolumn - 1] | occup[your_colour][icolumn + 1]);
		n |= (occup[your_colour][icolumn] | occup[my_colour][icolumn]);
		if (my_colour == WHITE_COLOUR)
			tst = (irow >= bit_high[n]);
		else
			tst = (irow <= bit_low[n]);
		if (tst)
		{
			if (verbose)
				my_printf("%spawn on %s is free\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
			sum += free_pawn_grad[my_colour][p->board_where];
			continue;
		}
		
		/*
		We have an isolated pawn if there are no pawns to the left and to the right.
		*/
		
		if (icolumn == 0)
			n = occup[my_colour][1];
		else if (icolumn == 7)
			n = occup[my_colour][6];
		else
			n = (occup[my_colour][icolumn - 1] | occup[my_colour][icolumn + 1]);
		if (n == 0)
		{
			if (verbose)
				my_printf("%spawn on %s is isolated\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
			sum += isolated_pawn_score;
			pawn_board[p->board_where] = TRUE;
			continue;
		}
		
		/*
		This is just a poor check for a lagged pawn.
		*/
		
		if (my_colour == WHITE_COLOUR)
			tst = (irow < (bit_low[n] - 1));
		else
			tst = (irow > (bit_high[n] + 1));
		if (tst)
		{
			if (verbose)
				my_printf("%spawn on %s is lagged\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
			sum += lagged_pawn_score;
			pawn_board[p->board_where] = TRUE;
		}
	}
	return(sum);
}

/*
return_pos_score analyzes the remaining pieces. Since the attack board
is generated only once (for the colour to move), we have to check either >= 1 
or <= -1. This is controlled by the flag inverse).
*/

local int return_pos_score(int, int, int, int);
local int return_pos_score(int my_colour, int your_colour, int verbose, int inverse)
{
#define MY_PAWN(X) ((board[X]->board_colour == my_colour) and (board[X]->board_kind == PAWN))
#define MY_ROOK(X) ((board[X]->board_colour == my_colour) and (board[X]->board_kind == ROOK))
	board_t *p, *q;
	int tst;
	int icolumn, irow;
	int n, m;
	int sum = 0;
	int j, k;
	int *d;
	
	for (p = knight_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		sum += knight_visit[my_colour][p->board_where];
		if (!p->board_moved) sum += knight_not_developed_score;
		irow = row[p->board_where];
		icolumn = column[p->board_where];
		
		/*
		A knight is strong if it is a 'free pawn' and covered by one or 
		(even better) two pawns
		*/
		
		if (icolumn == 0)
			n = occup[your_colour][1];
		else if (icolumn == 7)
			n = occup[your_colour][6];
		else
			n = (occup[your_colour][icolumn - 1] | occup[your_colour][icolumn + 1]);
		if (my_colour == WHITE_COLOUR)
			tst = ((n == 0) or (irow >= bit_high[n]));
		else
			tst = ((n == 0) or (irow <= bit_low[n]));
		if (tst)
		{
			q = board[p->board_where - pawn_dir[my_colour][1]];
			if ((q->board_colour == my_colour) and (q->board_kind == PAWN))
			{
				if (verbose)
					my_printf("%sknight on %s is strong\n",
						colour_as_string[my_colour], GET_NOTA(p->board_where));
				sum += knight_strong_score;
			}
			q = board[p->board_where - pawn_dir[my_colour][2]];
			if ((q->board_colour == my_colour) and (q->board_kind == PAWN))
			{
				if (verbose)
					my_printf("%sknight on %s is strong\n",
						colour_as_string[my_colour], GET_NOTA(p->board_where));
				sum += knight_strong_score;
			}
		}
		
		/*
		We define mobility as the number of squares to which the knight can
		move without the risk of being captured. n is the number of 
		squares controlled by my_colour, m is the number of squares controlled
		by your_colour.
		Note: we have to check if attack_board >= 1, not >= 0 because the piece
		itself is counted when generating the attack board.
		*/
		
		n = m = 0;
		for (d = p->board_dir; *d != 0; d++)
		{
			j = p->board_where + *d;
			q = board[j];
			if (q == &empty)
			{
				if (inverse)
					tst = (attack_board[j] <= -1);
				else
					tst = (attack_board[j] >= 1);
				if (tst)
					n += knight_can_move_score;
				else
					m += knight_can_move_score;
			}
			else if (q->board_colour == your_colour)
			{
				if (q->board_kind == PAWN)
				{
					if (pawn_board[j])
					{
						if (verbose)
							my_printf("%sknight on %s attacks weak pawn on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
						sum += weak_pawn_attack_score;
					}
				}
			}
		}
		
		/*
		A knight is weak if it can only move to sqares controlled by the opponent.
		*/
		
		if ((n == 0) and (m > 0))
		{
			if (verbose)
				my_printf("%sknight on %s is weak\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
			sum += weak_knight_score;
		}
		else
			sum += (n - m);
	}
	
	/*
	Same for bishop and rook.
	*/
	
	for (p = bishop_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		if (!p->board_moved) sum += bishop_not_developed_score;
		irow = row[p->board_where];
		icolumn = column[p->board_where];
		j = p->board_where - pawn_dir[my_colour][0];
		q = board[j];
		if ((q->board_colour == my_colour) and (q->board_kind == PAWN))
		{
			if (verbose)
				my_printf("%sbishop on %s blocks pawn on %s\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
			sum += bishop_blocks_pawn_score;
		}
		if (icolumn == 0)
			n = occup[your_colour][1];
		else if (icolumn == 7)
			n = occup[your_colour][6];
		else
			n = (occup[your_colour][icolumn - 1] | occup[your_colour][icolumn + 1]);
		if (my_colour == WHITE_COLOUR)
			tst = ((n == 0) or (irow >= bit_high[n]));
		else
			tst = ((n == 0) or (irow <= bit_low[n]));
		if (tst)
		{
			q = board[p->board_where - pawn_dir[my_colour][1]];
			if ((q->board_colour == my_colour) and (q->board_kind == PAWN))
			{
				if (verbose)
					my_printf("%sbishop on %s is strong\n",
						colour_as_string[my_colour], GET_NOTA(p->board_where));
				sum += bishop_strong_score;
			}
			q = board[p->board_where - pawn_dir[my_colour][2]];
			if ((q->board_colour == my_colour) and (q->board_kind == PAWN))
			{
				if (verbose)
					my_printf("%sbishop on %s is strong\n",
						colour_as_string[my_colour], GET_NOTA(p->board_where));
				sum += bishop_strong_score;
			}
		}
		n = m = 0;
		for (d = p->board_dir; *d != 0; d++)
		{
			for (j = p->board_where + *d; board[j] == &empty; j += *d)
			{
				if (inverse)
					tst = (attack_board[j] <= -1);
				else
					tst = (attack_board[j] >= 1);
				if (tst)
					n += bishop_can_move_score;
				else
					m += bishop_can_move_score;
			}
			q = board[j];
			if (q->board_colour == your_colour)
			{
				if (q->board_kind == PAWN)
				{
					if (pawn_board[j])
					{
						if (verbose)
							my_printf("%sbishop on %s attacks weak pawn on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
						sum += weak_pawn_attack_score;
					}
				}
				else
				{
					for (k = q->board_where + *d; board[k] == &empty; k += *d);
					q = board[k];
					if ((q->board_colour == your_colour) and (q->board_kind == KING))
					{
						if (verbose)
							my_printf("%sbishop on %s pins piece on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
						sum += pin_score;
					}
				}
			}
		}
		if ((n == 0) and (m > 0))
		{
			if (verbose)
				my_printf("%sbishop on %s is weak\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
			sum += weak_bishop_score;
		}
		else
			sum += (n - m);
	}	
	for (p = rook_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		irow = row[p->board_where];
		icolumn = column[p->board_where];
		n = (occup[my_colour][icolumn] | occup[your_colour][icolumn]);
		if (n == 0)
		{
			if (verbose)
				my_printf("%srook on %s is on open file\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
			sum += rook_on_open_file_score;
		}
		else if ((occup[my_colour][icolumn] == 0) or 
		         (occup[your_colour][icolumn] == 0))
		{
			if (verbose)
				my_printf("%srook on %s is on half open file\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
			sum += rook_on_half_open_file_score;
		}
		if (irow == strong_row[my_colour])
		{
			if (verbose)
				my_printf("%srook on %s is on 7th rank\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
			sum += rook_on_7th_rank_score;
		}
		n = m = 0;
		for (d = p->board_dir; *d != 0; d++)
		{
			for (j = p->board_where + *d; board[j] == &empty; j += *d)
			{
				if (inverse)
					tst = (attack_board[j] <= -1);
				else
					tst = (attack_board[j] >= 1);
				if (tst)
					n += rook_can_move_score;
				else
					m += rook_can_move_score;
			}
			q = board[j];
			if (q->board_colour == your_colour)
			{
				if (q->board_kind == PAWN)
				{
					if (pawn_board[j])
					{
						if (verbose)
							my_printf("%srook on %s attacks weak pawn on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
						sum += weak_pawn_attack_score;
					}
				}
				else
				{
					for (k = q->board_where + *d; board[k] == &empty; k += *d);
					q = board[k];
					if ((q->board_colour == your_colour) and (q->board_kind == KING))
					{
						if (verbose)
							my_printf("%srook on %s pins piece on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
						sum += pin_score;
					}
				}
			}
		}
		if ((n == 0) and (m > 0))
		{
			if (verbose)
				my_printf("%sweak rook on %s\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
			sum += weak_rook_score;
		}
		else
			sum += (n - m);
	}
	for (p = queen_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		n = m = 0;
		for (d = p->board_dir; *d != 0; d++)
		{
			for (j = p->board_where + *d; board[j] == &empty; j += *d)
			{
				if (inverse)
					tst = (attack_board[j] <= -1);
				else
					tst = (attack_board[j] >= 1);
				if (tst)
					n++;
				else
					m++;
			}
			q = board[j];
			if (q->board_colour == your_colour)
			{
				if (q->board_kind == PAWN)
				{
					if (pawn_board[j])
					{
						if (verbose)
							my_printf("%squeen on %s attacks weak pawn on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
						sum += weak_pawn_attack_score;
					}
				}
				else
				{
					for (k = q->board_where + *d; board[k] == &empty; k += *d);
					q = board[k];
					if ((q->board_colour == your_colour) and (q->board_kind == KING))
					{
						if (verbose)
							my_printf("%squeen on %s pins piece on %s\n",
								colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
						sum += pin_score;
					}
				}
			}
		}
		if ((n == 0) and (m > 0))
		{
			if (verbose)
				my_printf("%sweak queen on %s\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where));
			sum += weak_queen_score;
		}
	}
	
	/*
	King safety
	*/
	
	p = king_list[my_colour];
	for (d = p->board_dir; *d != 0; d++)
	{
		j = p->board_where + *d;
		if (board[j] == &side) continue;
		if (inverse)
			tst = (attack_board[j] > 0);
		else
			tst = (attack_board[j] < 0);
		if (tst)
		{
			if (verbose)
				my_printf("%sking on %s is in danger on %s\n",
					colour_as_string[my_colour], GET_NOTA(p->board_where), GET_NOTA(j));
			sum += king_danger_score;
		}
	}
	
	/*
	We assume castling is important if the opponent has at least a
	queen and a rook
	*/
	
	if ((piece_count[your_colour][QUEEN] > 0) and (piece_count[your_colour][ROOK] > 0))
	{
		/*
		Very poor castle check
		*/
		
		if (my_colour == WHITE_COLOUR)
		{
			if ((p->board_where == G1) or (p->board_where == H1) or
			    (p->board_where == G2) or (p->board_where == H2))
			{
				n = m = 0;
				if (MY_PAWN(F2)) n++;
				if (MY_PAWN(G2)) n++;
				if (MY_PAWN(H2)) n++;
				if (MY_PAWN(F3)) n++;
				if (MY_PAWN(G3)) n++;
				if (MY_PAWN(H3)) n++;
				if (MY_ROOK(G1)) m++;
				if (MY_ROOK(H1)) m++;
				if (MY_ROOK(G2)) m++;
				if (MY_ROOK(H2)) m++;
				if ((m == 0) and (n > 2))
				{
					if (verbose)
						my_printf("%sking on %s has short castled\n",
							colour_as_string[my_colour], GET_NOTA(p->board_where));
					sum += short_castle_score;
				}
			}
			else if ((p->board_where == C1) or
			         (p->board_where == B1) or (p->board_where == A1) or
			         (p->board_where == B2) or (p->board_where == A2))
			{
				n = m = 0;
				if (MY_PAWN(C2)) n++;
				if (MY_PAWN(B2)) n++;
				if (MY_PAWN(A2)) n++;
				if (MY_PAWN(C3)) n++;
				if (MY_PAWN(B3)) n++;
				if (MY_PAWN(A3)) n++;
				if (MY_ROOK(B1)) m++;
				if (MY_ROOK(A1)) m++;
				if (MY_ROOK(B2)) m++;
				if (MY_ROOK(A2)) m++;
				if ((m == 0) and (n > 2))
				{
					if (verbose)
						my_printf("%sking on %s has long castled\n",
							colour_as_string[my_colour], GET_NOTA(p->board_where));
					sum += long_castle_score;
				}
			}
		}
		else
		{
			if ((p->board_where == G8) or (p->board_where == H8) or
			    (p->board_where == G7) or (p->board_where == H7))
			{
				n = m = 0;
				if (MY_PAWN(F7)) n++;
				if (MY_PAWN(G7)) n++;
				if (MY_PAWN(H7)) n++;
				if (MY_PAWN(F6)) n++;
				if (MY_PAWN(G6)) n++;
				if (MY_PAWN(H6)) n++;
				if (MY_ROOK(G8)) m++;
				if (MY_ROOK(H8)) m++;
				if (MY_ROOK(G7)) m++;
				if (MY_ROOK(H7)) m++;
				if ((m == 0) and (n > 2))
				{
					if (verbose)
						my_printf("%sking on %s has short castled\n",
							colour_as_string[my_colour], GET_NOTA(p->board_where));
					sum += short_castle_score;
				}
			}
			else if ((p->board_where == C8) or
			         (p->board_where == B8) or (p->board_where == A8) or
			         (p->board_where == B7) or (p->board_where == A7))
			{
				n = m = 0;
				if (MY_PAWN(C7)) n++;
				if (MY_PAWN(B7)) n++;
				if (MY_PAWN(A7)) n++;
				if (MY_PAWN(C6)) n++;
				if (MY_PAWN(B6)) n++;
				if (MY_PAWN(A6)) n++;
				if (MY_ROOK(B8)) m++;
				if (MY_ROOK(A8)) m++;
				if (MY_ROOK(B7)) m++;
				if (MY_ROOK(A7)) m++;
				if ((m == 0) and (n > 2))
				{
					if (verbose)
						my_printf("%sking on %s has long castled\n",
							colour_as_string[my_colour], GET_NOTA(p->board_where));
					sum += long_castle_score;
				}
			}
		}
	}
	return(sum);
#undef MY_PAWN
#undef MY_ROOK
}

void display_score(void)
{
	int nwhite_mat, nblack_mat;
	int nwhite_pawn, nblack_pawn;
	int nwhite_pos, nblack_pos;
	int nwhite_tot, nblack_tot;
	int ndiff, nendgame;
	
	if ((tot_count[WHITE_COLOUR] + tot_count[BLACK_COLOUR]) < 4)
	{
		nendgame = return_endgame_score(WHITE_COLOUR, BLACK_COLOUR);
		my_printf("endgame=%d\n", nendgame);
		return;
	}
	
	set_attack_board(WHITE_COLOUR, BLACK_COLOUR);
	nwhite_mat = return_mat_score(WHITE_COLOUR);
	nblack_mat = return_mat_score(BLACK_COLOUR);
	nwhite_pawn = return_pawn_pos_score(WHITE_COLOUR, BLACK_COLOUR, TRUE);
	nblack_pawn = return_pawn_pos_score(BLACK_COLOUR, WHITE_COLOUR, TRUE);
	nwhite_pos = return_pos_score(WHITE_COLOUR, BLACK_COLOUR, TRUE, FALSE);
	nblack_pos = return_pos_score(BLACK_COLOUR, WHITE_COLOUR, TRUE, TRUE);
	nwhite_tot = nwhite_pawn + nwhite_pos;		
	nblack_tot = nblack_pawn + nblack_pos;

	my_printf("mat score white=%d black=%d\n", nwhite_mat, nblack_mat);
	my_printf("pawn score white=%d black=%d\n", nwhite_pawn, nblack_pawn);
	my_printf("pos score white=%d black=%d\n", nwhite_pos, nblack_pos);
	my_printf("tot pos score white=%d black=%d\n", nwhite_tot, nblack_tot);
	ndiff = nwhite_mat - nblack_mat + nwhite_tot - nblack_tot;
	my_printf("diff score=%d\n", ndiff);
}

/*
We check for draw by repetition by checking if the current key has occurred
before. In the alpha beta loop, we return TRUE when we find a match.
In the main loop, we must check whether the position has occurred three times.
*/

int draw_by_repetition(int strict)
{
	node_t *temp_node;
	int ncount;
	
  temp_node = node;
  ncount = 1;
  while(TRUE)
  {
  	/*
  	your move and my move must be reversible.
  	*/
  	
  	if (!temp_node->node_reverse) break;
  	temp_node--;
  	if (!temp_node->node_reverse) break;
  	temp_node--;
  	if (EQUAL_KEY(node->node_total_key, temp_node->node_total_key))
  	{
  		ncount++;
  		if (!strict)
  			return(TRUE);
  		else if (ncount > 2)
  			return(TRUE);
  	}
  }
  return(FALSE);
}

/*
heapsort routine. Sorts the moves m by a.
A straight implementation in C of the algorithm described by Knuth in
The art of computer programming, Volume 3, pages 146-147.
The arrays should be offset one.

local void heapsort(int n, int *, move_t *);
local void heapsort(int n, int *a, move_t *m)
{
	int l, r, b, i, j;
	move_t t;

	if (n < 2) return;
	l = n / 2 + 1;
	r = n;
	h2:
	if (l > 1)
	{
		l--;
		b = a[l]; t = m[l];
	}
	else
	{
		b = a[r]; t = m[r];
		a[r] = a[1]; m[r] = m[1];
		r--;
		if (r == 1)
		{
			a[1] = b; m[1] = t;
			goto h9;
		}		
	}
	j = l;
	h4:
	i = j;
	j = 2 * j;
	if (j == r) goto h6;
	if (j > r) goto h8;
	if (a[j] < a[j + 1]) j++;
	h6:
	if (b >= a[j]) goto h8;
	a[i] = a[j]; m[i] = m[j];
	goto h4;
	h8:
	a[i] = b; m[i] = t;
	goto h2;
	h9:
	for (i = 1; i < n; i++)
		BUG(a[i] > a[i + 1])
}

*/

#define QUIES_DEPTH 2

/*
A difficult problem in any chess program is the quiescense search (the capture
evaluation). See for example
G. Schrufer, ICCA vol. 12, no. 1 (1989) pp. 3-9.
I do it in the following way: any piece that has captured is marked.
In the quiescense search, I allow QUIES_DEPTH free captures (so any piece
can capture any other piece), but after the QUIES_DEPTH free captures, only
captures to pieces which have their capture mark set are allowed. In this way the
quiescence search will always terminate. Of course higher QUIES_DEPTH values
give better results, but also makes the program slower, so there is a trade-off
here.
*/

local score_t quies(int, score_t, score_t, int, var_t *);
local score_t quies(int swap, score_t alpha, score_t beta, int depth, var_t *var_base)
{
	score_t temp_score, best_score;
#if MACHINE == DOS
	score_cache_t huge *p;
#else
	score_cache_t *p;
#endif
	int nmoves;
	int i;
	int ikind;
	board_t *q;
	var_t *var, *var_top;
	move_t *base;
	
#ifdef BUG_DEPTH
	if (inode > inode_max)
	{
		int itemp;
		my_printf("inode max=%d\n", inode - inode_root);
		for (itemp = inode_root; itemp < inode; itemp++)
		{
			move_to_string(node_stack[itemp].node_move);
			my_printf(" %s", move_string);
		}
		my_printf("\n");
		inode_max = inode;
	}
#endif

	var_top = var_base;
	*var_top++ = INVAL_MOVE;
	
	if (swap)
	{
		temp_score = alpha;
		alpha = -beta;
		beta = -temp_score;
	}
	
	base = capt_top;

	best_score = node->node_mat[my_colour] - node->node_mat[your_colour];
	if (UsePos == 0) goto label_skip;
	
	if ((tot_count[WHITE_COLOUR] + tot_count[BLACK_COLOUR]) < 4)
	{
		best_score += return_endgame_score(my_colour, your_colour);
		goto label_skip;
	}
	p = score_cache[my_colour] + (node->node_total_key.low & SCORE_CACHE_MASK);
	if ((p->cache_score != SCORE_INVAL) and
	    (EQUAL_KEY(p->cache_key, node->node_total_key)))
	{
#ifdef CACHE_FEN
		char fen_string[FEN_STRING_MAX];
		board_to_fen(fen_string, FALSE);
		BUG(strcmp(fen_string, p->cache_fen) != 0)
#endif
		best_score += p->cache_score;
		nhits++;
	}
	else
	{
		set_attack_board(my_colour, your_colour);
		temp_score = 0;
		temp_score += return_pawn_pos_score(my_colour, your_colour, FALSE);
		temp_score -= return_pawn_pos_score(your_colour, my_colour, FALSE);
		temp_score += return_pos_score(my_colour, your_colour, FALSE, FALSE);
		temp_score -= return_pos_score(your_colour, my_colour, FALSE, TRUE);
		
		best_score += temp_score;
		nevals++;
		p->cache_score = temp_score;
		p->cache_key = node->node_total_key;
#ifdef CACHE_FEN
		board_to_fen(p->cache_fen, FALSE);
#endif
	}
	
	/*
	If by doing nothing we already have a fail high, we do not have to 
	check the captures.
	*/
	
	label_skip:
	if (best_score >= beta)
		goto label_neg;

	/*
	The captures. We generate the captures to your most valuable pieces first.
	*/
	
	for (ikind = QUEEN; ikind >= PAWN; ikind--)
	{
		for (q = list[your_colour][ikind]; q->board_kind != INVALID; q++)
		{
			if (q->board_where == INVALID) continue;
			
			/*
			Only recaptures if (depth >= QUIES_DEPTH)
			*/
			
			if ((depth >= QUIES_DEPTH) and !q->board_captured) continue;
			capt_top = base;
			nmoves = gen_capts(q->board_where, my_colour);
			for (i = 0; i < nmoves; i++)
			{
				do_move(base[i]);
				if (i_am_check(your_colour, my_colour))
				{
					undo_move();
					continue;
				}
				temp_score = quies(TRUE, MAX(alpha, best_score), beta, depth + 1, var_top);
				undo_move();
				if (temp_score > best_score)
				{
					best_score = temp_score;
					var = var_base;
					*var++ = base[i];
					while ((*var++ = *var_top++) != INVAL_MOVE);
					var_top = var;
					var = var_base;
					if (best_score >= beta) goto label_neg;
				}
			}
		}
	}
	
	label_neg:
	capt_top = base;
	if (swap)	
		best_score = -best_score;
	return(best_score);
}

/*
A standard alpha beta routine, with the null move heuristic. See for example
T.A. Marsland, ICCA vol. 9, no. 1 (1986) pp. 3-19.
For the null move heuristic see 
Chr. Donninger, ICCA vol. 12, no 3 (1993) pp. 137-143.
Null moves are used as a forward pruning mechanism: if the null move gives
a beta cutoff, it is assumed that the regular moves in that position, when searched,
would also give a beta cutoff. This speeds up the program by a factor of two, but
it fails terribly in case of zugzwang, because in that case not doing a move is a 
usually a good idea: the null move gives a beta cutoff, but the regular moves will not.
I call this 'null move failures'.
It is very easy to check how often null move failures occur: instead of
doing the null move cutoff, you set a flag. If you do not get a beta cutoff
when searching the regular moves and that flag is set, you know you have 
a null move failure. I tried this for the 300 win-at-chess positions 
from Lang&Smith, searched with DepthMax = 5. In 5 out of the 300 positions,
I got null move failure warnings. It still found the right solution
in 4 of the 5 positions, and in 1 position it did not find the solution any more.
*/

#ifdef PVM
local int quit_tag[PVM_PROCS_MAX] = {PVM_INVALID_TAG};
local int nquit_tag = 0;
#endif

#if DEBUG_LEVEL > 0
local score_t debug_alpha_beta(int, score_t, score_t, int, var_t *);
local score_t debug_alpha_beta(int null_move, score_t alpha, score_t beta, int depth, var_t *var_base)
{
#define RETURN(X)\
	{best_score = (X); goto label_neg;}
	
	score_t best_score, temp_score;
	int best_move;
	int check;
	int nmoves, nlegal;
	int imove, jmove;
	int depth_next;
	var_t *var, *var_top;
	move_t *base;
	
	node_t *node_debug = node;
#if MACHINE == MAC
	if (Button()) return(SCORE_INVAL);
#endif

	BUG(alpha >= beta)
	BUG(depth < 0)
	
	var_top = var_base;
	*var_top++ = INVAL_MOVE;
	
	clock_end = clock();
	if (clock_end < clock_begin)
	{
		my_printf("clock time cycle detected!\n");
		clock_begin = clock_end;
	}
	else
	{
		if ((clock_end - clock_begin) > (CLOCKS_PER_SEC / 10))
		{
			time_used += (double) (clock_end - clock_begin) / CLOCKS_PER_SEC;
			clock_begin = clock_end;
			if (time_used > TimeLimit) return(SCORE_INVAL);
		}
	}
	
	temp_score = alpha;
	alpha = -beta;
	beta = -temp_score;
	
	base = capt_top;

	best_score = -SCORE_LIMIT;
	best_move = INVALID;
	
	if (draw_by_repetition(FALSE)) RETURN(0)
	
	check = i_am_check(my_colour, your_colour);
	
	if (!check and (depth == 0))
	{
#ifdef BUG_DEPTH
		bug_depth[inode - inode_root]++;
#endif
		best_score = quies(FALSE, alpha, beta, 0, var_base);
		goto label_neg;
	}

	if (!check and !null_move and
	    (depth >= NULL_MOVE_DEPTH) and (node->node_mat[my_colour] >= (2*ROOK_SCORE)))
	{
		do_move(NULL_MOVE);
		depth_next = depth - 1;
		temp_score = debug_alpha_beta(TRUE, beta - 1, beta, depth_next - NULL_MOVE_REDUC, var_top);
		undo_move();
		if (temp_score == SCORE_INVAL) return(SCORE_INVAL);
		if (temp_score >= beta) RETURN(temp_score)
	}

	nlegal = 0;
	if (check)
		depth_next = depth;
	else
		depth_next = depth - 1;

	ngens++;
	nmoves = gen_all_moves();
	
	if (best_move == INVALID)
		best_move = 0;
	else if (best_move >= nmoves)
	{
		my_printf("hash fault!\n");
		best_move = 0;
	}

	imove = INVALID;
	while(imove < nmoves)
	{
		if (imove == INVALID)
			jmove = best_move;
		else
			jmove = imove;
		if (base[jmove] == INVAL_MOVE)
		{
			imove++;
			continue;
		}
		do_move(base[jmove]);
		if (i_am_check(your_colour, my_colour))
		{
			undo_move();
			base[jmove] = INVAL_MOVE;
			imove++;
			continue;
		}
		nlegal++;
		ndos++;
		temp_score = debug_alpha_beta(FALSE, MAX(alpha, best_score), beta, depth_next, var_top);
		undo_move();
		if (temp_score == SCORE_INVAL) 
			return(SCORE_INVAL);
		if (temp_score > best_score)
		{
			best_score = temp_score;
			best_move = jmove;
			var = var_base;
			*var++ = base[jmove];
			while ((*var++ = *var_top++) != INVAL_MOVE);
			var_top = var;
			var = var_base;
			if (best_score >= beta) goto label_neg;
		}
		base[jmove] = INVAL_MOVE;
		imove++;
	}
	
	if (nlegal == 0)
	{
		if (check)
			RETURN(-SCORE_MATE + inode)
		else
			RETURN(0)
	}

	label_neg:
	BUG(best_score == -SCORE_LIMIT)
	BUG(node != node_debug)

	capt_top = base;
	return(-best_score);
#undef RETURN
}
#endif

local score_t alpha_beta(int, score_t, score_t, int, var_t *);
local score_t alpha_beta(int null_move, score_t alpha, score_t beta, int depth, var_t *var_base)
{
#define RETURN(X)\
	{best_score = (X); goto label_neg;}
	
	score_t best_score, temp_score;
#if MACHINE == DOS
	cache_t huge *p;
#else
	cache_t *p;
#endif
	int best_move;
	int check;
	int nmoves, nlegal;
	int imove, jmove;
	int depth_next;
	var_t *var, *var_top;
	move_t *base;
	
#if DEBUG_LEVEL > 0
	node_t *node_debug = node;
	int debug_score;
#endif
#ifdef PVM
	char fen_string[LINE_MAX];
	int familyid;
	int bufid;
	int all_done;
	int ntotal, nmaster, nslave;
	int has_slave;
#endif
#if MACHINE == MAC
	if (Button()) return(SCORE_INVAL);
#endif

	BUG(alpha >= beta)
	BUG(depth < 0)
	
	var_top = var_base;
	*var_top++ = INVAL_MOVE;
	
	/*
	A previous version of my program occasionally stuck
	in infinite loops when run on an RS6000. After some debugging,
	I found out that it had to do with the time control:
	The code looked something like:
	t1 = clock();
	..
	t2 = clock();
	if ((t2 - t1) > (TimeLimit * CLOCKS_PER_SEC)) return..
	
	Looks good, but only works on machines with a slow clock (like my MAC).
	On an RS6000 CLOCKS_PER_SEC is 1000000L, and clock_t is a long,
	so the clock cycles in about 40 minutes. Now suppose t1 is LONG_MAX (or
	near LONG_MAX), then t2 is always less than t1 and the program
	will not terminate!
	To overcome this problem, I now update the time used every 1/10 of 
	a second. When a clock cycle is detected a warning is printed (no attempt
	is made to recover from the clock cycle, so the time used may by off
	by at most 1/10 of a second).
	*/
	
	clock_end = clock();
	if (clock_end < clock_begin)
	{
		my_printf("clock time cycle detected!\n");
		clock_begin = clock_end;
	}
	else
	{
		if ((clock_end - clock_begin) > (CLOCKS_PER_SEC / 10))
		{
			time_used += (double) (clock_end - clock_begin) / CLOCKS_PER_SEC;
			clock_begin = clock_end;
#ifdef PVM
			PVMBUG(bufid = pvm_nrecv(-1, quit_tag[nquit_tag]))
			if (bufid > 0) return(SCORE_INVAL);
#else
			if (time_used > TimeLimit) return(SCORE_INVAL);
#endif
		}
	}
	
	/*
	I swap alpha and beta here to keep the call to alpha_beta readable.
	*/
	
	temp_score = alpha;
	alpha = -beta;
	beta = -temp_score;
	
	base = capt_top;

	best_score = -SCORE_LIMIT;
	best_move = INVALID;
	
	/*
	We first check the hash table. Since we use the lower CACHE_KIND_BITS bits
	to store some information, we do not check these bits.
	*/
	
	p = cache[my_colour] + (node->node_total_key.low & CACHE_MASK);
	if (!(p->cache_key.low & CACHE_INVAL) and
	    (p->cache_key.high == node->node_total_key.high) and
#ifdef HASH_KEY128
	    (p->cache_key.lower == node->node_total_key.lower) and
	    (p->cache_key.higher == node->node_total_key.higher) and
#endif
	    ((p->cache_key.low & ~CACHE_KIND_MASK) == (node->node_total_key.low & ~CACHE_KIND_MASK)))
	{
#ifdef CACHE_FEN
		char fen_string[FEN_STRING_MAX];
		board_to_fen(fen_string, FALSE);
		BUG(strcmp(fen_string, p->cache_fen) != 0)
#endif
		best_move = p->cache_best;
		
		/*
		Checking if cache_depth >= depth is possibly dangerous.
		We certainly need to check cache_depth == depth if DEBUG_LEVEL > 1,
		otherwise the comparison with debug_alpha_beta fails.
		*/
		
#if DEBUG_LEVEL > 1
		if ((p->cache_key.low & CACHE_DEPTH_MASK) == depth)
#else
		if ((p->cache_key.low & CACHE_DEPTH_MASK) >= depth)
#endif	
		{
			if (p->cache_key.low & CACHE_TRUE)
				RETURN(p->cache_score)
			else if (p->cache_key.low & CACHE_LOWER)
			{
				if (p->cache_score >= beta)
					RETURN(p->cache_score)
				else if (p->cache_score > alpha)
					alpha = p->cache_score;
			}
			else if (p->cache_key.low & CACHE_UPPER)
			{
				if (p->cache_score <= alpha)
					RETURN(p->cache_score)
				else if (p->cache_score < beta)
					beta = p->cache_score;
			}
		}
	}
	
	if (draw_by_repetition(FALSE)) RETURN(0)
	
	check = i_am_check(my_colour, your_colour);
	
	/*
	We have an end node if we are not check and depth == 0.
	*/
	
	if (!check and (depth == 0))
	{
#ifdef BUG_DEPTH
		bug_depth[inode - inode_root]++;
#endif
		best_score = quies(FALSE, alpha, beta, 0, var_base);
		goto label_neg;
	}

	/*
	Try the null move. We assume no zugzwang if material >= (2*ROOK_SCORE).
	*/
	
	if (!check and !null_move and
	    (depth >= NULL_MOVE_DEPTH) and (node->node_mat[my_colour] >= (2*ROOK_SCORE)))
	{
		do_move(NULL_MOVE);
		depth_next = depth - 1;
		temp_score = alpha_beta(TRUE, beta - 1, beta, depth_next - NULL_MOVE_REDUC, var_top);
		undo_move();
		if (temp_score == SCORE_INVAL) return(SCORE_INVAL);
		if (temp_score >= beta) RETURN(temp_score)
	}

	/*
	Since we generate pseudo legal moves, we have to keep track of 
	the number of legal moves.
	*/
	
	nlegal = 0;
	if (check)
		depth_next = depth;
	else
		depth_next = depth - 1;

	/*
	Generate all moves.
	*/
	
	ngens++;
	nmoves = gen_all_moves();
	
	/*
	Search the moves, the move from the hash table first.
	*/
	
	if (best_move == INVALID)
		best_move = 0;
	else if (best_move >= nmoves)
	{
		my_printf("hash fault!\n");
		best_move = 0;
	}

#ifdef PVM
	familyid = INVALID;
	all_done = FALSE;
#endif
	imove = INVALID;
	while(imove < nmoves)
	{
#ifdef PVM
		if (familyid != INVALID)
		{
			PVMBUG(bufid = pvm_nrecv(-1, familyid))
			if (bufid > 0)
			{
				PVMBUG(pvm_upkint(&temp_score, 1, 1))
				BUG(temp_score == SCORE_INVAL)
				if (temp_score > best_score)
				{
					best_score = temp_score;
					if (best_score >= beta)
						goto label_store;
				}
			}
		}
#endif
	
		if (imove == INVALID)
			jmove = best_move;
		else
			jmove = imove;
		if (base[jmove] == INVAL_MOVE)
		{
			imove++;
			continue;
		}
		do_move(base[jmove]);
		if (i_am_check(your_colour, my_colour))
		{
			undo_move();
			base[jmove] = INVAL_MOVE;
			imove++;
			continue;
		}
		nlegal++;
		ndos++;
		temp_score = alpha_beta(FALSE, MAX(alpha, best_score), beta, depth_next, var_top);
		undo_move();
#ifdef PVM
		if (temp_score == SCORE_INVAL)
		{
			if (familyid == INVALID)
				return(SCORE_INVAL);
			else
				goto label_store;
		}
#else
		if (temp_score == SCORE_INVAL) 
			return(SCORE_INVAL);
#endif
		if (temp_score > best_score)
		{
#ifdef PVM
			if (familyid != INVALID)
			{
				PVMBUG(pvm_initsend(PvmDataDefault))
				PVMBUG(pvm_pkint(&familyid, 1, 1))
				PVMBUG(pvm_pkint(&temp_score, 1, 1))
				PVMBUG(pvm_send(drivertid, PVM_UPDATE_TAG))
			}
#endif
			best_score = temp_score;
			best_move = jmove;
			var = var_base;
			*var++ = base[jmove];
			while ((*var++ = *var_top++) != INVAL_MOVE);
			var_top = var;
			var = var_base;
			if (best_score >= beta) goto label_store;
		}
		base[jmove] = INVAL_MOVE;
		imove++;
#ifdef PVM
		if (check or null_move or (depth < PvmDepth)) continue;
		if (nlegal < PvmLegal) continue;
		if ((imove % PvmLegal) != 0) continue;
		ntotal = nmoves - imove;
		if (ntotal <= PvmLegal) continue;
		
		/*
		Divide moves between master and slave.
		The master should always analyze at least PvmLegal moves,
		the slave might analyze fewer than PvmLegal moves to hopefully
		reduce the time the master has to wait.
		*/
		
		nmaster = PvmLegal;
		nslave = ntotal - nmaster;
		if (nslave > PvmLegal) nslave = PvmLegal;
		
		board_to_fen(fen_string, FALSE);
		PVMBUG(pvm_initsend(PvmDataDefault))
		PVMBUG(pvm_pkint(&familyid, 1, 1))
		PVMBUG(pvm_pkstr(fen_string))
		PVMBUG(pvm_pkint(&depth, 1, 1))
		PVMBUG(pvm_pkint(&nslave, 1, 1))
		PVMBUG(pvm_pklong(base + imove, nslave, 1))
		if (familyid == INVALID)
		{
			PVMBUG(pvm_pkint(&best_score, 1, 1))
			PVMBUG(pvm_pkint(&alpha, 1, 1))
			PVMBUG(pvm_pkint(&beta, 1, 1))
		}
		PVMBUG(pvm_send(drivertid, PVM_ALLOCATE_TAG))

		PVMBUG(pvm_recv(-1, PVM_ALLOCATE_TAG))
		PVMBUG(pvm_upkint(&has_slave, 1, 1))
		if (!has_slave) continue;
		
		/*
		We have a slave.
		*/
		
		if (familyid == INVALID)
		{
			PVMBUG(pvm_upkint(&familyid, 1, 1))
			nquit_tag++;
			BUG(nquit_tag >= PVM_PROCS_MAX)
			PVMBUG(pvm_upkint(quit_tag + nquit_tag, 1, 1))
		}
		imove += nslave;
#endif
	}
	
	/*
	When detecting check mate, we add inode to the mate score
	to make sure we find the shortest mate.
	*/
	
	if (nlegal == 0)
	{
		if (check)
			RETURN(-SCORE_MATE + inode)
		else
			RETURN(0)
	}

	label_store:
#ifdef PVM
	if (familyid != INVALID)
	{
		PVMBUG(pvm_initsend(PvmDataDefault))
		PVMBUG(pvm_pkint(&familyid, 1, 1))
		PVMBUG(pvm_send(drivertid, PVM_WAITING_TAG))
		while(!all_done)
		{
			PVMBUG(bufid = pvm_recv(-1, familyid))
			PVMBUG(pvm_upkint(&temp_score, 1, 1))
			if (temp_score == SCORE_INVAL)
				all_done = TRUE;
			else if (temp_score > best_score)
				best_score = temp_score;
		}
		nquit_tag--;
		if (options.print_distrib > 0)
		{
			display_board();
			my_printf("familyid=%d alpha=%d beta=%d best_score=%d\n",
				familyid, alpha, beta, best_score);
		}
	}
#endif
#if DEBUG_LEVEL > 1

	/*
	Sometimes inverting scores is nice, sometimes it just isn't..
	*/
	
	debug_score = debug_alpha_beta(FALSE, -beta, -alpha, depth, var_top);
	BUG(debug_score == SCORE_INVAL)
	debug_score = -debug_score;
	if (best_score >= beta)
	{
		if (debug_score < beta)
		{
			my_printf("alpha=%d beta=%d best_score=%d debug_score=%d\n",
				alpha, beta, best_score, debug_score);
			FATAL_ERROR("debug_score < beta")
		}
	}
	else if (best_score > alpha)
	{
		if (debug_score >= beta)
		{
			my_printf("alpha=%d beta=%d best_score=%d debug_score=%d\n",
				alpha, beta, best_score, debug_score);
			FATAL_ERROR("debug_score >= beta")
		}
		else if (debug_score < alpha)
		{
			my_printf("alpha=%d beta=%d best_score=%d debug_score=%d\n",
				alpha, beta, best_score, debug_score);
			FATAL_ERROR("debug_score < alpha")
		}
	}
	else
	{
		if (debug_score > alpha)
		{
			my_printf("alpha=%d beta=%d best_score=%d debug_score=%d\n",
				alpha, beta, best_score, debug_score);
			FATAL_ERROR("debug_score > alpha")
		}
	}
#endif

#ifdef PVM
	if (familyid != INVALID) goto label_neg;
#endif

	if (depth > CACHE_DEPTH_MASK) goto label_neg;
	p->cache_key.high = node->node_total_key.high;
	
	/*
	We stuff the depth and the flags in the lower CACHE_KIND_BITS.
	*/
	
	p->cache_key.low = ((node->node_total_key.low & ~CACHE_KIND_MASK) | depth);
	p->cache_score = best_score;
	if (best_score <= alpha)
		p->cache_key.low |= CACHE_UPPER;
	else if (best_score >= beta)
		p->cache_key.low |= CACHE_LOWER;
	else
		p->cache_key.low |= CACHE_TRUE;
	p->cache_best = best_move;
#ifdef CACHE_FEN
	board_to_fen(p->cache_fen, FALSE);
#endif

	label_neg:
	BUG(best_score == -SCORE_LIMIT)
#if DEBUG_LEVEL > 0
	BUG(node != node_debug)
#endif

	capt_top = base;
	return(-best_score);
#undef RETURN
}

#ifdef PVM
void do_slave(void)
{
	int islave;
	char name[FILENAME_MAX];
	int mytid;
	int familyid;
	int bufid;
	char fen_string[LINE_MAX];
	int nmoves;
	move_t moves[MOVE_MAX];
	int best_score;
	int imove;
	int temp_score;
	int alpha, beta;
	int depth;
	time_t wall1, wall2;
	double wall;
	int nbytes, tag, tid;
	
	/*
	Enroll in pvm.
	*/
	
	wall1 = time(NULL);
	time_used = 0;
	
	PVMBUG(mytid = pvm_mytid())
	familyid = INVALID;
	
	PVMBUG(bufid = pvm_recv(-1, PVM_NUMBER_TAG))
	PVMBUG(pvm_upkint(&islave, 1, 1))
	PVMBUG(pvm_upkint(&drivertid, 1, 1))
	sprintf(name, "slave%d.%d.out", islave, mytid);
	my_printf("%z", name);

	/*
	A slave waits for an allocate.
	*/
	
	while(TRUE)
	{
		PVMBUG(bufid = pvm_recv(-1, -1))
		PVMBUG(pvm_bufinfo(bufid, &nbytes, &tag, &tid))
		if (tag == PVM_QUIT_TAG)
		{
			wall2 = time(NULL);
			wall = difftime(wall2, wall1);
			PVMBUG(pvm_initsend(PvmDataDefault))
			PVMBUG(pvm_pkdouble(&time_used, 1, 1))
			PVMBUG(pvm_pkdouble(&wall, 1, 1))
			PVMBUG(pvm_send(drivertid, PVM_QUIT_TAG))
			PVMBUG(pvm_exit())
			exit(EXIT_SUCCESS);
		}
		else if (tag == PVM_ALLOCATE_TAG)
		{
			PVMBUG(pvm_upkint(&familyid, 1, 1))
			nquit_tag++;
			BUG(nquit_tag >= PVM_PROCS_MAX)
			PVMBUG(pvm_upkint(quit_tag + nquit_tag, 1, 1))
			PVMBUG(pvm_upkstr(fen_string))
			PVMBUG(pvm_upkint(&depth, 1, 1))
			PVMBUG(pvm_upkint(&nmoves, 1, 1))
			PVMBUG(pvm_upklong(moves, nmoves, 1))
			PVMBUG(pvm_upkint(&best_score, 1, 1))
			PVMBUG(pvm_upkint(&alpha, 1, 1))
			PVMBUG(pvm_upkint(&beta, 1, 1))
			fen_to_board(fen_string, FALSE);
			
			for (imove = 0; imove < nmoves; imove++)
			{
				/*
				Check to see if better score is available.
				*/
			
				PVMBUG(bufid = pvm_nrecv(-1, familyid))
				if (bufid > 0)
				{
					PVMBUG(pvm_upkint(&temp_score, 1, 1))
					BUG(temp_score == SCORE_INVAL)
					if (temp_score > best_score)
					{
						best_score = temp_score;
						if (best_score >= beta)
							goto label_wait;
					}
				}
				
				if (moves[imove] == INVAL_MOVE) continue;
				do_move(moves[imove]);
				if (i_am_check(your_colour, my_colour))
				{
					undo_move();
					continue;
				}
				temp_score = alpha_beta(FALSE, MAX(alpha, best_score), beta, depth - 1, var_stack);
				undo_move();
				if (temp_score == SCORE_INVAL)
					goto label_wait;
				if (temp_score > best_score)
				{
					PVMBUG(pvm_initsend(PvmDataDefault))
					PVMBUG(pvm_pkint(&familyid, 1, 1))
					PVMBUG(pvm_pkint(&temp_score, 1, 1))
					PVMBUG(pvm_send(drivertid, PVM_UPDATE_TAG))
					best_score = temp_score;
				}
				if (best_score >= beta)
					break;
			}
			label_wait:
			if (options.print_distrib > 0)
			{
				display_board();
				my_printf("familyid=%d alpha=%d beta=%d best_score=%d\n",
					familyid, alpha, beta, best_score);
			}
			PVMBUG(pvm_initsend(PvmDataDefault))
			PVMBUG(pvm_pkint(&familyid, 1, 1))
			PVMBUG(pvm_send(drivertid, PVM_IDLE_TAG))
			while(TRUE)
			{
				PVMBUG(bufid = pvm_recv(-1, familyid))
				PVMBUG(pvm_upkint(&temp_score, 1, 1))
				if (temp_score == SCORE_INVAL) break;
			}
			nquit_tag--;
		}
		else
		{
			my_printf("ignoring %d\n", tag);
		}
	}
}
#endif

/*
mate_alpha_beta is a special alpha beta routine for mate finding
*/

local score_t mate_alpha_beta(score_t, score_t, int, var_t *);
local score_t mate_alpha_beta(score_t alpha, score_t beta, int depth, var_t *var_base)
{
#define RETURN(X)\
	{best_score = (X); goto label_neg;}
	
	score_t best_score, temp_score;
#if MACHINE == DOS
	cache_t huge *p;
#else
	cache_t *p;
#endif
	int best_move;
	int check;
	int nmoves, nlegal;
	int imove, jmove;
	var_t *var, *var_top;
	move_t *base;
	
#if MACHINE == MAC
	if (Button()) return(SCORE_INVAL);
#endif
	
	BUG(alpha >= beta)
	BUG(depth < 0)
	
	var_top = var_base;
	*var_top++ = INVAL_MOVE;
	
	clock_end = clock();
	if (clock_end < clock_begin)
	{
		my_printf("clock time cycle detected!\n");
		clock_begin = clock_end;
	}
	else
	{
		if ((clock_end - clock_begin) > (CLOCKS_PER_SEC / 10))
		{
			time_used += (double) (clock_end - clock_begin) / CLOCKS_PER_SEC;
			clock_begin = clock_end;
			if (time_used > TimeLimit) return(SCORE_INVAL);
		}
	}
	
	temp_score = alpha;
	alpha = -beta;
	beta = -temp_score;
	
	base = capt_top;

	best_score = -SCORE_LIMIT;
	best_move = INVALID;
	
	p = cache[my_colour] + (node->node_total_key.low & CACHE_MASK);
	if (!(p->cache_key.low & CACHE_INVAL) and
	    (p->cache_key.high == node->node_total_key.high) and
#ifdef HASH_KEY128
	    (p->cache_key.lower == node->node_total_key.lower) and
	    (p->cache_key.higher == node->node_total_key.higher) and
#endif
	    ((p->cache_key.low & ~CACHE_KIND_MASK) == (node->node_total_key.low & ~CACHE_KIND_MASK)))
	{
		best_move = p->cache_best;
		if ((p->cache_key.low & CACHE_DEPTH_MASK) >= depth)
		{
			if (p->cache_key.low & CACHE_TRUE)
				RETURN(p->cache_score)
			else if (p->cache_key.low & CACHE_LOWER)
			{
				if (p->cache_score >= beta)
					RETURN(p->cache_score)
				else if (p->cache_score > alpha)
					alpha = p->cache_score;
			}
			else if (p->cache_key.low & CACHE_UPPER)
			{
				if (p->cache_score <= alpha)
					RETURN(p->cache_score)
				else if (p->cache_score < beta)
					beta = p->cache_score;
			}
		}
	}
	
	if (draw_by_repetition(FALSE)) RETURN(0)
	
	check = i_am_check(my_colour, your_colour);
	
	if ((depth == 0) and !check)
	{
		best_score = 0;
		goto label_store;
	}
	
	nlegal = 0;
	ngens++;
	nmoves = gen_all_moves();
	
	if (depth == 0)
	{
		for (imove = 0; imove < nmoves; imove++)
		{
			do_move(base[imove]);
			if (i_am_check(your_colour, my_colour))
			{
				undo_move();
				continue;
			}
			else
			{
				undo_move();
				best_score = 0;
				goto label_store;
			}
		}
		best_score = -SCORE_MATE + inode;
		goto label_store;
	}
	
	if (best_move == INVALID)
		best_move = 0;
	else if (best_move >= nmoves)
	{
		my_printf("hash fault!\n");
		best_move = 0;
	}
	
	for (imove = INVALID; imove < nmoves; imove++)
	{
		if (imove == INVALID)
			jmove = best_move;
		else
			jmove = imove;
		if (base[jmove] == INVAL_MOVE) continue;
		do_move(base[jmove]);
		if (i_am_check(your_colour, my_colour))
		{
			undo_move();
			base[jmove] = INVAL_MOVE;
			continue;
		}
		ndos++;
		nlegal++;
		
		/*
		Another short cut. The move at depth 1 has to be a checking move.
		*/
		
		if ((depth == 1) and !i_am_check(my_colour, your_colour))
			temp_score = 0;
		else
			temp_score = mate_alpha_beta(MAX(alpha, best_score), beta, depth - 1, var_top);
		undo_move();
		if (temp_score == SCORE_INVAL) return(SCORE_INVAL);
		if (temp_score > best_score)
		{
			best_score = temp_score;
			best_move = jmove;
			var = var_base;
			*var++ = base[jmove];
			while ((*var++ = *var_top++) != INVAL_MOVE);
			var_top = var;
			var = var_base;
			if (best_score >= beta) goto label_store;
		}
		base[jmove] = INVAL_MOVE;
	}
	
	if (nlegal == 0)
	{
		if (check)
			RETURN(-SCORE_MATE + inode)
		else
			RETURN(0)
	}
	
	label_store:
	
	if (depth > CACHE_DEPTH_MASK) goto label_neg;
	p->cache_key.high = node->node_total_key.high;
	
	p->cache_key.low = ((node->node_total_key.low & ~CACHE_KIND_MASK) | depth);
	p->cache_score = best_score;
	if (best_score <= alpha)
		p->cache_key.low |= CACHE_UPPER;
	else if (best_score >= beta)
		p->cache_key.low |= CACHE_LOWER;
	else
		p->cache_key.low |= CACHE_TRUE;
	p->cache_best = best_move;
	
	label_neg:
	BUG(best_score == -SCORE_LIMIT)

	capt_top = base;
	return(-best_score);
#undef RETURN
}

/*
The legal moves in the starting position are stored in array root.
*/

int nroot;
root_t *root;
int jroot;

/*
gen_root finds the legal moves in the starting position.
*/

void gen_root(void)
{
	int imove, nmoves;
	int check;
	move_t *base;
				
	nroot = 0;
	capt_top = capt_stack;
	base = capt_top;
	nmoves = gen_all_moves();
	for (imove = 0; imove < nmoves; imove++)
	{
		do_move(base[imove]);
		check = i_am_check(your_colour, my_colour);
		undo_move();
		if (check) continue;
		BUG(nroot >= NROOT_MAX)
		root[nroot].root_move = base[imove];
		nroot++;
	}
	for (imove = 0; imove < nroot; imove++)
		root[imove].root_score = 0;
	jroot = 0;
}

/*
search_root checks whether move occurs in the list of root moves.
Returns INVALID if the move is not found.
Prints the legal moves if the move is not found and the verbose flag is set.
*/

int search_root(char *move, int option, int strict, int verbose)
{
	int iroot;
	int len;
	
	len = strlen(move);
	for (iroot = 0; iroot < nroot; iroot++)
	{
		if (option != MOVE_STRING_LONG)
		{
			move_to_san_string(root[iroot].root_move, TRUE);
			if (strict)
			{
				if (strcmp(move, move_string) == 0) break;
			}
			else
			{
				if (strncmp(move_string, move, len) == 0) break;
			}
		}
		if (option != MOVE_STRING_SAN)
		{
			move_to_string(root[iroot].root_move);
			if (strict)
			{
				if (strcmp(move, move_string) == 0) break;
			}
			else
			{
				if (strncmp(move_string, move, len) == 0) break;
			}
		}
	}
	if (iroot < nroot) return(iroot);
	if (verbose)
	{
		display_board();
		my_printf("move=%s\n", move);
		for (iroot = 0; iroot < nroot; iroot++)
		{
			move_to_san_string(root[iroot].root_move, TRUE);
			my_printf(" %7s", move_string);
			if (((iroot % 9) == 8) or (iroot == (nroot - 1))) my_printf("\n");
		}
	}
	return(INVALID);
}

/*
display_var displays the best variation.
*/

local void display_var(move_t, var_t *);
local void display_var(move_t best_move, var_t *var_base)
{
	move_to_san_string(best_move, TRUE);
	my_printf(" %s", move_string);
	if (draw_by_repetition(FALSE))
	{
		my_printf(" rep!");
		return;
	}
	do_move(best_move);
	if (i_am_check(your_colour, my_colour))
	{
		my_printf("hash fault, check!\n");
		display_board();
		move_to_string(best_move);
		my_printf("%s%s\n", colour_as_string[my_colour], move_string);
	}
	else
	{
		if (*var_base != INVAL_MOVE)
			display_var(*var_base, var_base + 1);
	}
	undo_move();
}

local void clear_cache(void);
local void clear_cache(void)
{
	int icolour;
	long i;

	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		for (i = 0; i < (CACHE_MASK + 1); i++)
			cache[icolour][i].cache_key.low = CACHE_INVAL;
		for (i = 0; i < (SCORE_CACHE_MASK + 1); i++)
			score_cache[icolour][i].cache_score = SCORE_INVAL;
	}
}

int get_book_move(void)
{
	int iroot;
	char *move;
	int i;

	if (search_dbase())
	{
		BUG(ndbase_records < 2)
		i = (int) (ulrand() % (ndbase_records - 1)) + 1;
		move = dbase_records[i];
		BUG((iroot = search_root(move, MOVE_STRING_ANY, TRUE, TRUE)) == INVALID)
		display_board();
		my_printf("book move:%s\n", move);
		return(iroot);
	}
	else
		return(INVALID);
}

/*
sort_root sorts the root moves by searching each move with a window of
-inf, inf and a depth of 1.
(not used yet).
*/

void sort_root(void)
{
	int iroot, kroot;
	root_t temp;
	
	if (ZapCache) clear_cache();
	inode_root = inode;
	inode_max = 0;
	
	clock_begin = clock();
	time_used = 0;
	TimeLimit = 100;
	
	ngens = ndos = 0;
	nevals = nhits = 0;

	for (iroot = 0; iroot < nroot; iroot++)
	{
		do_move(root[iroot].root_move);
		root[iroot].root_score = alpha_beta(FALSE, -SCORE_LIMIT, SCORE_LIMIT, 0, var_stack);
		undo_move();
		my_printf("ts");
		display_var(root[iroot].root_move, var_stack);
		my_printf(" %.3f\n", (double) root[iroot].root_score);
	}
	my_printf("time=%.2f stats=%ld/%ld/%ld/%ld\n",
		time_used, ngens, ndos, nevals, nhits);
	for (iroot = 0; iroot < (nroot - 1); iroot++)
	{
		jroot = iroot;
		for (kroot = iroot + 1; kroot < nroot; kroot++)
			if (root[kroot].root_score > root[jroot].root_score) jroot = kroot;
		if (jroot != iroot)
		{
			temp = root[iroot];
			root[iroot] = root[jroot];
			root[jroot] = temp;
		}
	}
	/*
	the best move is now the first move
	*/
	jroot = 0;
}

int do_mate(void)
{
	int iroot, depth;
	int temp_score;
	
	if (ZapCache) clear_cache();
	
	clock_begin = clock();
	time_used = 0;
	inode_root = inode;
	inode_max = 0;
	
	for (depth = 1; depth <= 2 * MateDepth - 1; depth += 2)
	{
		ngens = ndos = 0;
	
		for (iroot = 0; iroot < nroot; iroot++)
		{
			do_move(root[iroot].root_move);
			temp_score = mate_alpha_beta(SCORE_ABORT - 1, SCORE_ABORT, depth - 1, var_stack);
			undo_move();
			if (temp_score == SCORE_INVAL) return(FALSE);
			if (temp_score > SCORE_ABORT)
			{
				display_var(root[iroot].root_move, var_stack);
				my_printf(" %.3f\n", (double) temp_score);
				return(TRUE);
			}
		}
		my_printf("depth=%d time=%.2f stats=%ld/%ld\n",
			depth, time_used, ngens, ndos);
	}
	return(FALSE);
}

#define MARGIN (PAWN_SCORE / 2)

/*
do_root tries to find the best move by iterative deepening,
aspiration search and minimal window search.
*/

void do_root(void)
{
	int check;
	int depth;
	score_t temp_score, fh_score;
	int depth_next;
	score_t alpha, beta;
	int iroot;
	
	if (UseBook)
	{
		jroot = get_book_move();
		if (jroot != INVALID) return;
	}
	
	if (ZapCache) clear_cache();
	check = i_am_check(my_colour, your_colour);

	jroot = 0;
	root[0].root_score = node->node_mat[my_colour] - node->node_mat[your_colour];
	inode_root = inode;
	inode_max = 0;
	
	clock_begin = clock();
	time_used = 0;
	
	iroot = 0;
	temp_score = SCORE_INVAL;
	
	/*
	Iterative deepening.
	*/
	
	for (depth = 1; depth <= DepthMax; depth++)
	{
#ifdef BUG_DEPTH
		{
			int ibug;
			for (ibug = 0; ibug < NODE_MAX; ibug++)
				bug_depth[ibug] = 0;
		}
#endif
		
		depth_next = (check ? depth : depth - 1);
		ngens = ndos = 0;
		nevals = nhits = 0;
		
		/*
		Aspiration search with half a pawn margin.
		*/
		
		alpha = root[jroot].root_score - MARGIN;
		beta = root[jroot].root_score + MARGIN;
		
		do_move(root[jroot].root_move);
		temp_score = alpha_beta(FALSE, alpha, beta, depth_next, var_stack);
		undo_move();
		if (temp_score == SCORE_INVAL) goto label_final;
		
		/*
		Research if fail low or fail high.
		*/

		do_move(root[jroot].root_move);
		if (temp_score <= alpha)
			temp_score = alpha_beta(FALSE, -SCORE_LIMIT, temp_score, depth_next, var_stack);
		else if (temp_score >= beta)
			temp_score = alpha_beta(FALSE, temp_score, SCORE_LIMIT, depth_next, var_stack);
		undo_move();
		if (temp_score == SCORE_INVAL) goto label_final;
		root[jroot].root_score = temp_score;
		my_printf("pv %d", depth);
		display_var(root[jroot].root_move, var_stack);
		my_printf(" %.3f\n", (double) root[jroot].root_score);
		
		/*
		Search the remaining moves with a minimal window.
		*/
		
		for (iroot = 0; iroot < nroot; iroot++)
		{
			if (iroot == jroot) continue;

			alpha = root[jroot].root_score;
			beta = alpha + 1;
			
			do_move(root[iroot].root_move);
			temp_score = alpha_beta(FALSE, alpha, beta, depth_next, var_stack);
			undo_move();
			if (temp_score == SCORE_INVAL) goto label_final;
			
			/*
			Research if fail high.
			*/
			
			if (temp_score > beta)
			{
				jroot = iroot;
				root[jroot].root_score = temp_score;
				my_printf("fh %d", depth);
				display_var(root[jroot].root_move, var_stack);
				my_printf(" %.3f\n", (double) temp_score);
				do_move(root[jroot].root_move);
				
				fh_score = temp_score;
				temp_score = alpha_beta(FALSE, temp_score, SCORE_LIMIT, depth_next, var_stack);
				undo_move();
				if (temp_score == SCORE_INVAL) goto label_final;
				
				/*
				If the reseach fails low, there was a null move failure (alas),
				a (depth >= cache_depth) failure (see alpha beta) or a hash problem.
				*/
				
				if (temp_score < fh_score)
					my_printf("research error!\n");
				root[jroot].root_score = temp_score;
				my_printf("pv %d", depth);
				display_var(root[jroot].root_move, var_stack);
				my_printf(" %.3f\n", (double) root[jroot].root_score);
			}
		}
#ifdef BUG_DEPTH
		{
			int ibug;
			for (ibug = 0; ibug < NODE_MAX; ibug++)
				if (bug_depth[ibug] != 0) my_printf("%d %d\n", ibug, bug_depth[ibug]);
		}
#endif
		
		my_printf("depth=%d score=%.3f time=%.2f stats=%ld/%ld/%ld/%ld\n",
			depth, (double) root[jroot].root_score, time_used,
			ngens, ndos, nevals, nhits);
		if (ABS(root[jroot].root_score) > SCORE_ABORT) goto label_final;
	}
	label_final:
	
	/*
	Only print this if time limit.
	*/
	
	if (temp_score == SCORE_INVAL)
	{
		my_printf("time limit depth=%d move=%d time=%.2f stats=%ld/%ld/%ld/%ld\n",
			depth, iroot, time_used, ngens, ndos, nevals, nhits);
	}
	if (jroot == INVALID) jroot = 0;
	
#ifdef PVM
	pvm_initsend(PvmDataDefault);
	pvm_pkint(&dummy, 1, 1);
	pvm_send(drivertid, PVM_QUIT_TAG);
#endif
}

/*
Initialization of some arrays.
*/

void initialize(void)
{
	int i, j, n;
	int icolour, ikind;
	
	/*
	The bit count arrays.
	*/
	
	MALLOC(bit_count, int, NBITS_SIZE)
	MALLOC(bit_low, int, NBITS_SIZE)
	MALLOC(bit_high, int, NBITS_SIZE)
	for (i = 0; i < NBITS_SIZE; i++)
	{
		n = 0;
		for (j = 0; j < NBITS; j++)
			if ((i & BIT(j)) != 0) n++;
		bit_count[i] = n;
		if (i == 0)
			bit_low[i] = bit_high[i] = INVALID;
		else
		{
			for (j = 0; j < NBITS; j++)
				if ((i & BIT(j)) != 0) break;
			bit_low[i] = j;
			for (j = NBITS - 1; j >= 0; j--)
				if ((i & BIT(j)) != 0) break;
			bit_high[i] = j;
		}
	}
	MALLOC(row, int, BOARD_MAX)
	MALLOC(column, int, BOARD_MAX)
	for (i = 0; i < BOARD_MAX; i++)
		row[i] = column[i] = -1;
	for (i = 0; i < 64; i++)
	{
		row[map[i]] = row_table[i];
		column[map[i]] = column_table[i];
	}
	MALLOC(nota, char *, BOARD_MAX)
	for (i = 0; i < BOARD_MAX; i++)
		nota[i] = nota_inval;
	for (i = 0; i < 64; i++)
		nota[map[i]] = nota_table[i];
	MALLOC(move_stack, move_t, MOVE_MAX)
	MALLOC(capt_stack, move_t, CAPT_MAX)
	MALLOC(var_stack, var_t, VAR_MAX)
	move_top = move_stack;
	capt_top = capt_stack;
	MALLOC(node_stack, node_t, NODE_MAX)
	node_stack[0].node_reverse = FALSE;
	node_stack++;
	node = node_stack;
	inode = inode_max = 0;
	
	my_printf("size of hash tables: %lu %lu\n",
		(ulong_t) sizeof(cache_t) * CACHE_MASK * 2,
		(ulong_t) sizeof(score_cache_t) * SCORE_CACHE_MASK * 2);
#if MACHINE == DOS
	HUGE_MALLOC(cache[WHITE_COLOUR], cache_t, CACHE_MASK + 1)
	HUGE_MALLOC(cache[BLACK_COLOUR], cache_t, CACHE_MASK + 1)
	HUGE_MALLOC(score_cache[WHITE_COLOUR], score_cache_t, SCORE_CACHE_MASK + 1)
	HUGE_MALLOC(score_cache[BLACK_COLOUR], score_cache_t, SCORE_CACHE_MASK + 1)
#else
	MALLOC(cache[WHITE_COLOUR], cache_t, CACHE_MASK + 1)
	MALLOC(cache[BLACK_COLOUR], cache_t, CACHE_MASK + 1)
	MALLOC(score_cache[WHITE_COLOUR], score_cache_t, SCORE_CACHE_MASK + 1)
	MALLOC(score_cache[BLACK_COLOUR], score_cache_t, SCORE_CACHE_MASK + 1)
#endif
	clear_cache();

	/*
	The hash keys.
	*/
	
	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		for (ikind = PAWN; ikind <= KING; ikind++)
		{
#if MACHINE == DOS
			HUGE_MALLOC(hash_key[icolour][ikind], hash_key_t, BOARD_MAX)
#else
			MALLOC(hash_key[icolour][ikind], hash_key_t, BOARD_MAX)
#endif
			for (i = 0; i < 64; i++) SET_KEY(hash_key[icolour][ikind][map[i]])
		}
	}
#if MACHINE == DOS
	HUGE_MALLOC(ep_key, hash_key_t, BOARD_MAX)
	HUGE_MALLOC(castle_key, hash_key_t, BOARD_MAX)
#else
	MALLOC(ep_key, hash_key_t, BOARD_MAX)
	MALLOC(castle_key, hash_key_t, BOARD_MAX)
#endif
	for (i = 0; i < 64; i++)
	{
		SET_KEY(ep_key[map[i]])
		SET_KEY(castle_key[map[i]])
	}
	
	/*
	Some more tables.
	*/
	
	MALLOC(king_alone_grad, int, BOARD_MAX)
	MALLOC(white_free_pawn_grad, int, BOARD_MAX)
	MALLOC(black_free_pawn_grad, int, BOARD_MAX)
	free_pawn_grad[WHITE_COLOUR] = white_free_pawn_grad;
	free_pawn_grad[BLACK_COLOUR] = black_free_pawn_grad;

	MALLOC(white_pawn_visit, int, BOARD_MAX)
	MALLOC(black_pawn_visit, int, BOARD_MAX)
	pawn_visit[WHITE_COLOUR] = white_pawn_visit;
	pawn_visit[BLACK_COLOUR] = black_pawn_visit;

	MALLOC(white_knight_visit, int, BOARD_MAX)
	MALLOC(black_knight_visit, int, BOARD_MAX)
	knight_visit[WHITE_COLOUR] = white_knight_visit;
	knight_visit[BLACK_COLOUR] = black_knight_visit;

	MALLOC(pawn_board, int, BOARD_MAX)
	for (i = 0; i < BOARD_MAX; i++)
	{
		king_alone_grad[i] = 0;
		white_free_pawn_grad[i] = black_free_pawn_grad[i] = 0;
		white_pawn_visit[i] = black_pawn_visit[i] = 0;
		white_knight_visit[i] = black_knight_visit[i] = 0;
	}
	for (i = 0; i < 64; i++)
	{
		king_alone_grad[map[i]] = king_alone_grad_table[i];
		white_free_pawn_grad[map[i]] = free_pawn_grad_table[i];
		black_free_pawn_grad[mirror_map[i]] = free_pawn_grad_table[i];
		white_pawn_visit[map[i]] = pawn_visit_table[i];
		black_pawn_visit[mirror_map[i]] = pawn_visit_table[i];
		white_knight_visit[map[i]] = knight_visit_table[i];
		black_knight_visit[mirror_map[i]] = knight_visit_table[i];
	}
	
	/*
	Scale the free pawn gradient such that a pawn on rank 7 is worth a pawn
	*/
	
	for (i = 0; i < 64; i++)
	{
		white_free_pawn_grad[map[i]] *= 3;
		black_free_pawn_grad[map[i]] *= 3;
	}
	MALLOC(attack_board, int, BOARD_MAX)
	
	MALLOC(root, root_t, NROOT_MAX)
	
	if (options.random_opening)
	{
		n = time(NULL) % 1000;
		for (i = 0; i < n; i++) ulrand();
	}
}
